"""Test 3D backpropagation algorithm"""
import ctypes
import multiprocessing as mp
import platform
import sys

import numpy as np

import odtbrain
from odtbrain import _alg3d_bpp

from common_methods import create_test_sino_3d, cutout, \
    get_test_parameter_set, write_results, get_results

WRITE_RES = False


def helper_3d_backprop_phase():
    sino, angles = create_test_sino_3d()
    parameters = get_test_parameter_set(2)
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                      dtype=float, **p)
        r.append(cutout(f))
    data = np.array(r).flatten().view(float)
    return data


def test_3d_backprop_phase():
    myframe = sys._getframe()
    sino, angles = create_test_sino_3d()
    parameters = get_test_parameter_set(2)
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                      dtype=float, **p)
        r.append(cutout(f))
    if WRITE_RES:
        write_results(myframe, r)
    data = np.array(r).flatten().view(float)
    assert np.allclose(data, get_results(myframe))


def test_3d_backprop_nopadreal():
    """
    - no padding
    - only real result
    """
    platform.system = lambda: "Windows"
    myframe = sys._getframe()
    sino, angles = create_test_sino_3d()
    parameters = get_test_parameter_set(2)
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padding=(False, False),
                                      dtype=float, onlyreal=True, **p)
        r.append(cutout(f))
    if WRITE_RES:
        write_results(myframe, r)
    data = np.array(r).flatten().view(float)
    assert np.allclose(data, get_results(myframe))


def test_3d_backprop_windows():
    """
    We assume that we are not running these tests on windows.
    So we perform a test with fake windows to increase coverage.
    """
    datalin = helper_3d_backprop_phase()
    real_system = platform.system
    datawin = helper_3d_backprop_phase()
    platform.system = real_system
    assert np.allclose(datalin, datawin)


def test_3d_backprop_real():
    """
    Check if the real reconstruction matches the real part
    of the complex reconstruction.
    """
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    parameters = get_test_parameter_set(2)
    # complex
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                      dtype=float,
                                      onlyreal=False, **p)
        r.append(f)
    # real
    r2 = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                      dtype=float,
                                      onlyreal=True, **p)
        r2.append(f)
    assert np.allclose(np.array(r).real, np.array(r2))


def test_3d_backprop_phase32():
    sino, angles = create_test_sino_3d()
    parameters = get_test_parameter_set(2)
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles,
                                      dtype=np.float32,
                                      padval=0,
                                      **p)
        r.append(cutout(f))
    data32 = np.array(r).flatten().view(np.float32)
    data64 = helper_3d_backprop_phase()
    assert np.allclose(data32, data64, atol=6e-7, rtol=0)


def test_3d_mprotate():
    myframe = sys._getframe()
    ln = 10
    ln2 = 2*ln
    initial_array = np.arange(ln2**3).reshape((ln2, ln2, ln2))
    shared_array = mp.RawArray(ctypes.c_double, ln2 * ln2 * ln2)
    arr = np.frombuffer(shared_array).reshape(ln2, ln2, ln2)
    arr[:, :, :] = initial_array
    _alg3d_bpp.mprotate_dict["X"] = shared_array
    _alg3d_bpp.mprotate_dict["X_shape"] = (ln2, ln2, ln2)

    pool = mp.Pool(processes=mp.cpu_count(),
                   initializer=_alg3d_bpp._init_worker,
                   initargs=(shared_array, (ln2, ln2, ln2), np.dtype(float)))
    _alg3d_bpp._mprotate(2, ln, pool, 2)
    if WRITE_RES:
        write_results(myframe, arr)
    assert np.allclose(np.array(arr).flatten().view(
        float), get_results(myframe))



"""Miscellaneous methods for example data handling"""
import lzma
import os
import pathlib
import tarfile
import tempfile
import warnings
import zipfile

import numpy as np


datapath = pathlib.Path(__file__).parent / "data"
webloc = "https://github.com/RI-imaging/ODTbrain/raw/master/examples/data/"


def dl_file(url, dest, chunk_size=6553):
    """Download `url` to `dest`"""
    import urllib3
    http = urllib3.PoolManager()
    r = http.request('GET', url, preload_content=False)
    with dest.open('wb') as out:
        while True:
            data = r.read(chunk_size)
            if data is None or len(data) == 0:
                break
            out.write(data)
    r.release_conn()


def extract_lzma(path):
    """Extract an lzma file and return the temporary file name"""
    tlfile = pathlib.Path(path)
    # open lzma file
    with tlfile.open("rb") as td:
        data = lzma.decompress(td.read())
    # write temporary tar file
    fd, tmpname = tempfile.mkstemp(prefix="odt_ex_", suffix=".tar")
    with open(fd, "wb") as fo:
        fo.write(data)
    return tmpname


def get_file(fname, datapath=datapath):
    """Return path of an example data file

    Return the full path to an example data file name.
    If the file does not exist in the `datapath` directory,
    tries to download it from the ODTbrain GitHub repository.
    """
    # download location
    datapath = pathlib.Path(datapath)
    datapath.mkdir(parents=True, exist_ok=True)

    dlfile = datapath / fname
    if not dlfile.exists():
        print("Attempting to download file {} from {} to {}.".
              format(fname, webloc, datapath))
        try:
            dl_file(url=webloc+fname, dest=dlfile)
        except BaseException:
            warnings.warn("Download failed: {}".format(fname))
            raise
    return dlfile


def load_data(fname, **kwargs):
    """Load example data"""
    fname = get_file(fname)
    if fname.suffix == ".lzma":
        return load_tar_lzma_data(fname)
    elif fname.suffix == ".zip":
        return load_zip_data(fname, **kwargs)


def load_tar_lzma_data(tlfile):
    """Load example sinogram data from a .tar.lzma file"""
    tmpname = extract_lzma(tlfile)

    # open tar file
    fields_real = []
    fields_imag = []
    phantom = []
    parms = {}

    with tarfile.open(tmpname, "r") as t:
        members = t.getmembers()
        members.sort(key=lambda x: x.name)

        for m in members:
            n = m.name
            f = t.extractfile(m)
            if n.startswith("fdtd_info"):
                for ln in f.readlines():
                    ln = ln.decode()
                    if ln.count("=") == 1:
                        key, val = ln.split("=")
                        parms[key.strip()] = float(val.strip())
            elif n.startswith("phantom"):
                phantom.append(np.loadtxt(f))
            elif n.startswith("field"):
                if n.endswith("imag.txt"):
                    fields_imag.append(np.loadtxt(f))
                elif n.endswith("real.txt"):
                    fields_real.append(np.loadtxt(f))

    try:
        os.remove(tmpname)
    except OSError:
        pass

    phantom = np.array(phantom)
    sino = np.array(fields_real) + 1j * np.array(fields_imag)
    angles = np.linspace(0, 2 * np.pi, sino.shape[0], endpoint=False)

    return sino, angles, phantom, parms


def load_zip_data(zipname, f_sino_real, f_sino_imag,
                  f_angles=None, f_phantom=None, f_info=None):
    """Load example sinogram data from a .zip file"""
    ret = []
    with zipfile.ZipFile(str(zipname)) as arc:
        sino_real = np.loadtxt(arc.open(f_sino_real))
        sino_imag = np.loadtxt(arc.open(f_sino_imag))
        sino = sino_real + 1j * sino_imag
        ret.append(sino)
        if f_angles:
            angles = np.loadtxt(arc.open(f_angles))
            ret.append(angles)
        if f_phantom:
            phantom = np.loadtxt(arc.open(f_phantom))
            ret.append(phantom)
        if f_info:
            with arc.open(f_info) as info:
                cfg = {}
                for li in info.readlines():
                    li = li.decode()
                    if li.count("=") == 1:
                        key, val = li.split("=")
                        cfg[key.strip()] = float(val.strip())
            ret.append(cfg)
    return ret



import multiprocessing as mp

import numpy as np
import pyfftw
import scipy.ndimage as ndi


def apple_core_3d(shape, res, nm):
    r"""Return a binary array with the apple core in 3D

    Parameters
    ----------
    shape: list-like, length 3
        Shape of the reconstruction volume for which to compute
        the apple core; The second (y-) axis is assumed to be the
        axis of symmetry (according to ODTbrain standard notation)
    res: float
        Size of the vacuum wave length :math:`\lambda` in pixels
    nm: float
        Refractive index of the medium :math:`n_\mathrm{m}`

    Returns
    -------
    core: 3D ndarray
        The mask is `True` for positions within the apple core
    """
    km = (2 * np.pi * nm) / res

    lNx, lNy, lNz = shape

    if lNx != lNz:
        raise ValueError("`shape[0]` and `shape[2]` must be identical!")

    fx = np.fft.fftfreq(lNx).reshape(-1, 1, 1)
    fy = np.fft.fftfreq(lNy).reshape(1, -1, 1)
    fz = np.fft.fftfreq(lNz).reshape(1, 1, -1)

    ky = 2*np.pi * fy
    kxz = 2*np.pi * np.sqrt(fx**2 + fz**2)
    kr = 2*np.pi * np.sqrt(fx**2 + fy**2 + fz**2)

    # 1. initiate empy array
    core = np.zeros(shape, dtype=bool)

    # 2. fill core
    root = 2*km*kxz - kxz**2
    root[root < 0] = 0
    core[np.abs(ky) > np.sqrt(root)] = True

    # 3. remove enveloping sphere (resolution limit)
    core[kr > np.sqrt(2) * km] = False

    return core


def constraint_nn(data, mask=None, bg_shell=None):
    """Non-negativity constraint"""
    # No imaginary RI (no absorption)
    if np.iscomplexobj(data):
        data.imag[:] = 0
    # background medium shell
    if bg_shell is not None:
        data.real[bg_shell] = 0
    # Also remove outer shell
    spov = spillover_region(data.shape)
    data.real[spov] = 0

    lowri = data.real < 0
    if mask is not None:
        # honor given mask
        lowri *= mask
    data.real[lowri] = 0


def constraint_sh(data, mask=None, bg_shell=None):
    """Symmetric histogram background data constraint"""
    # No imaginary RI (no absorption)
    if np.iscomplexobj(data):
        data.imag[:] = 0

    # determine range of medium RI (using background support)
    spov = spillover_region(data.shape)
    if bg_shell is not None:
        spov |= bg_shell
    fmin = np.min(data.real[spov])
    fmax = np.max(data.real[spov])

    # center
    full_hist, full_edge = np.histogram(
        data.real, bins=100, range=(fmin, fmax))
    de = full_edge[1] - full_edge[0]
    full_f = full_edge[1:] - de/2
    # center index (actually we would expect f_c==0)
    idx_c = np.argmax(full_hist)
    # half-maximum indices
    idx_start = idx_c - count_to_half(full_hist[:idx_c][::-1])
    idx_end = idx_c + count_to_half(full_hist[idx_c:])
    # RI values outside
    below = (data.real > fmin) * (data.real < full_f[idx_start])
    above = (data.real > full_f[idx_end]) * (data.real < fmax)
    out = below | above
    if mask is not None:
        # honor given mask
        out *= mask
    # push RI values to zero
    data.real[out] *= .5

    if bg_shell is not None:
        # push known background data to zero
        data.real[bg_shell] *= .5


def correct(f, res, nm, method="nn", mask=None, bg_shell_width=None,
            enforce_envelope=0.95, max_iter=100, min_diff=.01,
            count=None, max_count=None):
    r"""Fill the missing apple core of the object function

    Parameters
    ----------
    f: 3D ndarray
        Complex objec function :math:`f(\mathbf{r})`
    res: float
        Size of the vacuum wave length :math:`\lambda` in pixels
    nm: float
        Refractive index of the medium :math:`n_\mathrm{m}` that
        surrounds the object in :math:`n(\mathbf{r})`
    method: str
        One of:

        - "nn": non-negativity constraint (`f > 0`). This method
          resembles classic missing apple core correction.
        - "sh": symmetric histogram constraint (background data in
          `f`). This method works well for sparse-gradient data (e.g.
          works better than "nn" for simulated data), but might result
          in stripe-like artifacts when applied to experimental data.

        The imaginary part of the refractive index is suppressed
        in both cases.
        Note that these constraints are soft, i.e. after the final
        inverse Fourier transform, the conditions might not be met.

    mask: 3D boolean ndarray, or None
        Optional, defines background region(s) used for enforcing
        `method`. If a boolean ndarray, the values set to `True` define
        the used background regions.
    bg_shell_width: float
        Optional, defines the width of an ellipsoid shell (outer radii
        matching image shape) that is used additionally for enforcing
        `method`.
    enforce_envelope: float in interval [0,1] or False
        Set the suppression factor for frequencies that are above
        the envelope function; disabled if set to False or 0
    max_iter: int
        Maximum number of iterations to perform
    min_diff: float
        Stopping criterion computed as the relative difference
        (relative to the first iteration `norm`) of the changes applied
        during the current iteration `cur_diff`:
        ``np.abs(cur_diff/norm) < min_diff``
    count: multiprocessing.Value
        May be used for tracking progress. At each iteration
        `count.value` is incremented by one.
    max_count: multiprocessing.Value
        May be used for tracking progress; is incremented initially.

    Notes
    -----
    Internally, the Fourier transform is performed with single-precision
    floating point values (complex64).
    """
    if enforce_envelope < 0 or enforce_envelope > 1:
        raise ValueError("`enforce_envelope` must be in interval [0, 1]")

    if max_count is not None:
        with max_count.get_lock():
            max_count.value += max_iter + 2

    # Location of the apple core
    core = apple_core_3d(shape=f.shape, res=res, nm=nm)

    if count is not None:
        with count.get_lock():
            count.value += 1

    data = pyfftw.empty_aligned(f.shape, dtype='complex64')
    ftdata = pyfftw.empty_aligned(f.shape, dtype='complex64')
    fftw_forw = pyfftw.FFTW(data, ftdata,
                            axes=(0, 1, 2),
                            direction="FFTW_FORWARD",
                            flags=["FFTW_MEASURE"],
                            threads=mp.cpu_count())
    # Note: input array `ftdata` is destroyed when invoking `fftw_back`
    fftw_back = pyfftw.FFTW(ftdata, data,
                            axes=(0, 1, 2),
                            direction="FFTW_BACKWARD",
                            flags=["FFTW_MEASURE"],
                            threads=mp.cpu_count())

    data.real[:] = f.real
    data.imag[:] = 0

    fftw_forw.execute()
    ftdata_orig = ftdata.copy()

    if count is not None:
        with count.get_lock():
            count.value += 1

    if enforce_envelope:
        # Envelope function of Fourier amplitude
        ftevlp = envelope_gauss(ftdata_orig, core)

    init_state = np.sum(np.abs(ftdata_orig[core])) / data.size
    prev_state = init_state

    if bg_shell_width is not None:
        bg_shell = ellipsoid_shell(data.shape, width=bg_shell_width)
    else:
        bg_shell = None

    for ii in range(max_iter):
        if method == "nn":
            # non-negativity
            constraint_nn(data=data, mask=mask, bg_shell=bg_shell)
        elif method == "sh":
            # symmetric histogram
            constraint_sh(data=data, mask=mask, bg_shell=bg_shell)

        # Go into Fourier domain
        fftw_forw.execute()
        if enforce_envelope:
            # Suppress large frequencies with the envelope
            high = np.abs(ftdata) > ftevlp
            ftdata[high] *= enforce_envelope

        if method == "sh":
            # update dc term
            ftdata_orig[0, 0, 0] = (ftdata_orig[0, 0, 0] + ftdata[0, 0, 0])/2
        # Enforce original data
        ftdata[~core] = ftdata_orig[~core]

        fftw_back.execute()
        data[:] /= fftw_forw.N

        if count is not None:
            with count.get_lock():
                count.value += 1

        cur_state = np.sum(np.abs(ftdata[core])) / data.size
        cur_diff = cur_state - prev_state
        if ii == 0:
            norm = cur_diff
        else:
            if np.abs(cur_diff/norm) < min_diff:
                break
        prev_state = cur_state

    if count is not None:
        with count.get_lock():
            # add skipped counts (due to stopping criterion)
            count.value += max_iter - ii - 1

    return data


def count_to_half(array):
    """Determination of half-initial value index

    Return first index at which array values decrease below 1/2 of
    the initial initial value `array[0]`.
    """
    num = 0
    for item in array[1:]:
        if item < array[0] / 2:
            break
        else:
            num += 1
    return num


def ellipsoid_shell(shape, width=20):
    """Return background ellipsoid shell"""
    spov_outer = spillover_region(shape, shell=0)
    spov_inner = spillover_region(shape, shell=width)
    reg = spov_outer ^ spov_inner
    return reg


def envelope_gauss(ftdata, core):
    r"""Compute a gaussian-filtered envelope, without apple core

    Parameters
    ----------
    ftdata: 3D ndarray
        Fourier transform of the object function data
        (zero frequency not shifted to center of array)
    core: 3D ndarray (same shape as ftdata)
        Apple core (as defined by :func:`apple_core_3d`)

    Returns
    -------
    envelope: 3D ndarray
        Envelope function in Fourier space
    """
    hull = np.abs(ftdata)
    hull[core] = np.nan  # label core data with nans
    # Fill the apple core region with data from known regions from
    # the other axes (we only need an estimate if the envelope, so
    # this is a very good estimation of the Fourier amplitudes).
    shx, shy, _ = hull.shape
    maxsh = max(shx, shy)
    dsh = abs(shy - shx) // 2

    # Determine the slice
    if shx > shy:
        theslice = (slice(0, shx),
                    slice(dsh, shy+dsh),
                    slice(0, shx))
    else:
        theslice = (slice(dsh, shx+dsh),
                    slice(0, shy),
                    slice(dsh, shx+dsh),
                    )
    # 1. Create padded versions of the arrays, because shx and shy
    # can be different and inserting a transposed array will not work.
    hull_pad = np.zeros((maxsh, maxsh, maxsh), dtype=float)
    hull_pad[theslice] = np.fft.fftshift(hull)
    core_pad = np.zeros((maxsh, maxsh, maxsh), dtype=bool)
    core_pad[theslice] = np.fft.fftshift(core)
    # 2. Fill values from other axes were data are missing.
    hull_pad[core_pad] = np.transpose(hull_pad, (1, 0, 2))[core_pad]
    # 3. Fill any remaining nan-values (due to different shape or tilt)
    # with nearest neighbors. Use a distance transform for nearest
    # neighbor interpolation.
    invalid = np.isnan(hull_pad)
    ind = ndi.distance_transform_edt(invalid,
                                     return_distances=False,
                                     return_indices=True)
    hull_pad[:] = hull_pad[tuple(ind)]
    # 4. Write the data back to the original array.
    hull[:] = np.fft.ifftshift(hull_pad[theslice])
    # Perform gaussian blurring (shift data to make it smooth)
    gauss = ndi.gaussian_filter(input=np.fft.fftshift(hull),
                                sigma=np.max(ftdata.shape)/100,
                                mode="constant",
                                cval=0,
                                truncate=4.0)
    # Shift back gauss
    shifted_gauss = np.fft.ifftshift(gauss)
    return shifted_gauss


def spillover_region(shape, shell=0):
    """Return boolean array for region outside ellipsoid"""
    x = np.arange(shape[0]).reshape(-1, 1, 1)
    y = np.arange(shape[1]).reshape(1, -1, 1)
    z = np.arange(shape[2]).reshape(1, 1, -1)

    cx, cy, cz = np.array(shape) / 2
    spov = (((x-cx)/(cx-shell))**2
            + ((y-cy)/(cy-shell))**2
            + ((z-cz)/(cz-shell))**2) > 1
    return spov



"""2D slow integration"""
import warnings

import numpy as np

from .warn import DataUndersampledWarning


def integrate_2d(uSin, angles, res, nm, lD=0, coords=None,
                 count=None, max_count=None, verbose=0):
    r"""(slow) 2D reconstruction with the Fourier diffraction theorem

    Two-dimensional diffraction tomography reconstruction
    algorithm for scattering of a plane wave
    :math:`u_0(\mathbf{r}) = u_0(x,z)`
    by a dielectric object with refractive index
    :math:`n(x,z)`.

    This function implements the solution by summation in real
    space, which is extremely slow.

    Parameters
    ----------
    uSin: (A,N) ndarray
        Two-dimensional sinogram of line recordings
        :math:`u_{\mathrm{B}, \phi_j}(x_\mathrm{D})`
        divided by the incident plane wave :math:`u_0(l_\mathrm{D})`
        measured at the detector.
    angles: (A,) ndarray
        Angular positions :math:`\phi_j` of `uSin` in radians.
    res: float
        Vacuum wavelength of the light :math:`\lambda` in pixels.
    nm: float
        Refractive index of the surrounding medium :math:`n_\mathrm{m}`.
    lD: float
        Distance from center of rotation to detector plane
        :math:`l_\mathrm{D}` in pixels.
    coords: None or (2,M) ndarray]
        Computes only the output image at these coordinates. This
        keyword is reserved for future versions and is not
        implemented yet.
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.
    verbose: int
        Increment to increase verbosity.

    Returns
    -------
    f: ndarray of shape (N,N), complex if `onlyreal` is `False`
        Reconstructed object function :math:`f(\mathbf{r})` as defined
        by the Helmholtz equation.
        :math:`f(x,z) =
        k_m^2 \left(\left(\frac{n(x,z)}{n_m}\right)^2 -1\right)`


    See Also
    --------
    backpropagate_2d: implementation by backprojection
    fourier_map_2d: implementation by Fourier interpolation
    odt_to_ri: conversion of the object function :math:`f(\mathbf{r})`
        to refractive index :math:`n(\mathbf{r})`


    Notes
    -----
    This method is not meant for production use. The computation time
    is very long and the reconstruction quality is bad. This function
    is included in the package, because of its educational value,
    exemplifying the backpropagation algorithm.

    Do not use the parameter `lD` in combination with the Rytov
    approximation - the propagation is not correctly described.
    Instead, numerically refocus the sinogram prior to converting
    it to Rytov data (using e.g. :func:`odtbrain.sinogram_as_rytov`)
    with a numerical focusing algorithm (available in the Python
    package :py:mod:`nrefocus`).
    """
    if coords is None:
        lx = uSin.shape[1]
        x = np.linspace(-lx/2, lx/2, lx, endpoint=False)
        xv, yv = np.meshgrid(x, x)
        coords = np.zeros((2, lx**2))
        coords[0, :] = xv.flat
        coords[1, :] = yv.flat

    if max_count is not None:
        max_count.value += coords.shape[1] + 1

    # Cut-Off frequency
    km = (2 * np.pi * nm) / res

    # Fourier transform of all uB's
    # In the script we used the unitary angular frequency (uaf) Fourier
    # Transform. The discrete Fourier transform is equivalent to the
    # unitary ordinary frequency (uof) Fourier transform.
    #
    # uof: f₁(ξ) = int f(x) exp(-2πi xξ)
    #
    # uaf: f₃(ω) = (2π)^(-n/2) int f(x) exp(-i ωx)
    #
    # f₁(ω/(2π)) = (2π)^(n/2) f₃(ω)
    # ω = 2πξ
    #
    # We have a one-dimensional (n=1) Fourier transform and UB in the
    # script is equivalent to f₃(ω). Because we are working with the
    # uaf, we divide by sqrt(2π) after computing the fft with the uof.
    #
    # We calculate the fourier transform of uB further below. This is
    # necessary for memory control.

    # Corresponding sample frequencies
    fx = np.fft.fftfreq(uSin[0].shape[0])  # 1D array
    # kx is a 1D array.
    kx = 2 * np.pi * fx

    # Undersampling/oversampling?
    # Determine if the resolution of the image is too low by looking
    # at the maximum value for kx. This is no comparison between
    # Nyquist and Rayleigh frequency.
    if np.max(kx**2) <= 2 * km**2:
        # Detector is not set up properly. Higher resolution
        # can be achieved.
        warnings.warn("Measurement data are undersampled.",
                      DataUndersampledWarning)
    else:
        raise NotImplementedError("Oversampled data not yet supported." +
                                  " Please rescale input data")

    # Differentials for integral
    dphi0 = 2 * np.pi / len(angles)
    dkx = kx[1] - kx[0]

    # We will later multiply with phi0.
    # Make sure we are using correct shapes
    kx = kx.reshape(1, kx.shape[0])

    # Low-pass filter:
    # less-than-or-equal would give us zero division error.
    filter_klp = (kx**2 < km**2)

    # a0 will be multiplied with kx
    # a0 = np.atleast_1d(a0)
    # a0 = a0.reshape(1,-1)

    # Create the integrand
    # Integrals over ϕ₀ [0,2π]; kx [-kₘ,kₘ]
    #   - double coverage factor 1/2 already included
    #   - unitary angular frequency to unitary ordinary frequency
    #     conversion performed in calculation of UB=FT(uB).
    #
    # f(r) = -i kₘ / ((2π)^(3/2) a₀)            (prefactor)
    #      * iint dϕ₀ dkx                       (prefactor)
    #      * |kx|                               (prefactor)
    #      * exp(-i kₘ M lD )                   (prefactor)
    #      * UBϕ₀(kx)                             (dependent on ϕ₀)
    #      * exp( i (kx t⊥ + kₘ (M - 1) s₀) r )   (dependent on ϕ₀ and r)
    #
    # (r and s₀ are vectors. In the last term we perform the dot-product)
    #
    # kₘM = sqrt( kₘ² - kx² )
    # t⊥  = (  cos(ϕ₀), sin(ϕ₀) )
    # s₀  = ( -sin(ϕ₀), cos(ϕ₀) )
    #
    #
    # everything that is not dependent on phi0:
    #
    # Filter M so there are no nans from the root
    M = 1. / km * np.sqrt((km**2 - kx**2) * filter_klp)
    prefactor = -1j * km / ((2 * np.pi)**(3. / 2))
    prefactor *= dphi0 * dkx
    # Also filter the prefactor, so nothing outside the required
    # low-pass contributes to the sum.
    prefactor *= np.abs(kx) * filter_klp
    # new in version 0.1.4:
    # We multiply by the factor (M-1) instead of just (M)
    # to take into account that we have a scattered
    # wave that is normalized by u0.
    prefactor *= np.exp(-1j * km * (M-1) * lD)

    # Initiate function f
    f = np.zeros(len(coords[0]), dtype=np.complex128)
    lenf = len(f)
    lenu0 = len(uSin[0])  # lenu0 = len(kx[0])

    # Initiate vector r that corresponds to calculating a value of f.
    r = np.zeros((2, 1, 1))

    # Everything is normal.
    # Get the angles ϕ₀.
    phi0 = angles.reshape(-1, 1)
    # Compute the Fourier transform of uB.
    # This is true: np.fft.fft(UB)[0] == np.fft.fft(UB[0])
    # because axis -1 is always used.
    #
    #
    # Furthermore, The notation in the our optical tomography script for
    # a wave propagating to the right is:
    #
    #    u0(x) = exp(ikx)
    #
    # However, in physics usually usethe other sign convention:
    #
    #    u0(x) = exp(-ikx)
    #
    # In order to be consisten with programs like Meep or our scattering
    # script for a dielectric cylinder, we want to use the latter sign
    # convention.
    # This is not a big problem. We only need to multiply the imaginary
    # part of the scattered wave by -1.
    UB = np.fft.fft(np.fft.ifftshift(uSin, axes=-1)) / np.sqrt(2 * np.pi)
    UBi = UB.reshape(len(angles), lenu0)

    if count is not None:
        count.value += 1

    for j in range(lenf):
        # Get r (We compute f(r) in this for-loop)
        r[0][:] = coords[0, j]  # x
        r[1][:] = coords[1, j]  # y

        # Integrand changes with r, so we have to create a new
        # array:
        integrand = prefactor * UBi

        # We save memory by directly applying the following to
        # the integrand:
        #
        # Vector along which we measured
        # s0 = np.zeros((2, phi0.shape[0], kx.shape[0]))
        # s0[0] = -np.sin(phi0)
        # s0[1] = +np.cos(phi0)

        # Vector perpendicular to s0
        # t_perp_kx = np.zeros((2, phi0.shape[0], kx.shape[1]))
        #
        # t_perp_kx[0] = kx*np.cos(phi0)
        # t_perp_kx[1] = kx*np.sin(phi0)

        #
        # term3 = np.exp(1j*np.sum(r*( t_perp_kx + (gamma-km)*s0 ), axis=0))
        # integrand* = term3
        #
        # Reminder:
        # f(r) = -i kₘ / ((2π)^(3/2) a₀)            (prefactor)
        #      * iint dϕ₀ dkx                       (prefactor)
        #      * |kx|                               (prefactor)
        #      * exp(-i kₘ M lD )                   (prefactor)
        #      * UB(kx)                             (dependent on ϕ₀)
        #      * exp( i (kx t⊥ + kₘ(M - 1) s₀) r )   (dependent on ϕ₀ and r)
        #
        # (r and s₀ are vectors. In the last term we perform the dot-product)
        #
        # kₘM = sqrt( kₘ² - kx² )
        # t⊥  = (  cos(ϕ₀), sin(ϕ₀) )
        # s₀  = ( -sin(ϕ₀), cos(ϕ₀) )
        integrand *= np.exp(1j * (
            r[0] * (kx * np.cos(phi0) - km * (M - 1) * np.sin(phi0)) +
            r[1] * (kx * np.sin(phi0) + km * (M - 1) * np.cos(phi0))))

        # Calculate the integral for the position r
        # integrand.sort()
        f[j] = np.sum(integrand)

        # free memory
        del integrand

        if count is not None:
            count.value += 1

    return f.reshape(lx, lx)



"""Tests 1D angular weights"""
import numpy as np

import odtbrain

from common_methods import create_test_sino_2d, get_test_parameter_set


def test_angle_offset():
    """
    Tests if things are still correct when there is a 2PI offset in the angles.
    """
    sino, angles = create_test_sino_2d()
    parameters = get_test_parameter_set(2)
    # reference
    r1 = []
    for p in parameters:
        f1 = odtbrain.backpropagate_2d(sino, angles, weight_angles=False, **p)
        r1.append(f1)
    # with offset
    angles[::2] += 2*np.pi*np.arange(angles[::2].shape[0])
    r2 = []
    for p in parameters:
        f2 = odtbrain.backpropagate_2d(sino, angles, weight_angles=False, **p)
        r2.append(f2)
    # with offset and weights
    r3 = []
    for p in parameters:
        f3 = odtbrain.backpropagate_2d(sino, angles, weight_angles=True, **p)
        r3.append(f3)
    assert np.allclose(np.array(r1).flatten().view(float),
                       np.array(r2).flatten().view(float))
    assert np.allclose(np.array(r2).flatten().view(float),
                       np.array(r3).flatten().view(float))


def test_angle_swap():
    """
    Test if everything still works, when angles are swapped.
    """
    sino, angles = create_test_sino_2d()
    # remove elements so that we can see that weighting works
    angles = angles[:-2]
    sino = sino[:-2, :]
    parameters = get_test_parameter_set(2)
    # reference
    r1 = []
    for p in parameters:
        f1 = odtbrain.backpropagate_2d(sino, angles, weight_angles=True, **p)
        r1.append(f1)
    # change order of angles
    order = np.argsort(angles % .5)
    angles = angles[order]
    sino = sino[order, :]
    r2 = []
    for p in parameters:
        f2 = odtbrain.backpropagate_2d(sino, angles, weight_angles=True, **p)
        r2.append(f2)
    assert np.allclose(np.array(r1).flatten().view(float),
                       np.array(r2).flatten().view(float))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""3D backpropagation with tilted axis of rotation: sphere coordinates"""
import numpy as np

import odtbrain
import odtbrain._alg3d_bppt


def test_simple_sphere():
    """simple geometrical tests"""
    angles = np.array([0, np.pi/2, np.pi])
    axes = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1], [1, 0, 1], [1, 1, 1]]

    results = []

    for tilted_axis in axes:
        angle_coords = odtbrain._alg3d_bppt.sphere_points_from_angles_and_tilt(
            angles, tilted_axis)
        results.append(angle_coords)

    s2 = 1/np.sqrt(2)
    correct = np.array([[[1, 0, 0], [1, 0, 0], [1, 0, 0]],
                        [[0, 0, 1], [1, 0, 0], [0, 0, -1]],
                        [[0, 0, 1], [0, 0, 1], [0, 0, 1]],
                        [[0, 0, 1], [s2, .5, .5], [0, 1, 0]],
                        [[s2, 0, s2], [s2, 0, s2], [s2, 0, s2]],
                        [[s2, 0, s2],
                         [0.87965281125489458, s2/3*2, 0.063156230327168605],
                         [s2/3, s2/3*4, s2/3]],
                        ])
    assert np.allclose(correct, np.array(results))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()

    import matplotlib.pylab as plt
    from mpl_toolkits.mplot3d import Axes3D  # noqa: F401
    from matplotlib.patches import FancyArrowPatch
    from mpl_toolkits.mplot3d import proj3d

    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
            self._verts3d = xs, ys, zs

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, _zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            FancyArrowPatch.draw(self, renderer)

    axes = [[0, 1, 0], [0, 1, 0.1], [0, 1, -1], [1, 0.1, 0]]
    colors = ["k", "blue", "red", "green"]
    angles = np.linspace(0, 2*np.pi, 100)

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    for i in range(len(axes)):
        tilted_axis = axes[i]
        color = colors[i]
        tilted_axis = np.array(tilted_axis)
        tilted_axis = tilted_axis/np.sqrt(np.sum(tilted_axis**2))

        angle_coords = odtbrain._alg3d_bppt.sphere_points_from_angles_and_tilt(
            angles, tilted_axis)

        u, v, w = tilted_axis
        a = Arrow3D([0, u], [0, v], [0, w], mutation_scale=20,
                    lw=1, arrowstyle="-|>", color=color)
        ax.add_artist(a)
        ax.scatter(angle_coords[:, 0], angle_coords[:, 1],
                   angle_coords[:, 2], c=color, marker='o')

    radius = 1
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.set_xlim(-radius*1.5, radius*1.5)
    ax.set_ylim(-radius*1.5, radius*1.5)
    ax.set_zlim(-radius*1.5, radius*1.5)
    plt.tight_layout()

    plt.show()



"""Test 2d backpropagation"""
import sys

import numpy as np
import odtbrain

from common_methods import create_test_sino_2d, cutout, \
    get_test_parameter_set, write_results, get_results

WRITE_RES = False


def test_2d_backprop_phase():
    myframe = sys._getframe()
    sino, angles = create_test_sino_2d()
    parameters = get_test_parameter_set(2)
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_2d(sino, angles, **p)
        r.append(cutout(f))
    if WRITE_RES:
        write_results(myframe, r)
    assert np.allclose(np.array(r).flatten().view(float), get_results(myframe))


def test_2d_backprop_full():
    myframe = sys._getframe()
    sino, angles = create_test_sino_2d(ampl_range=(0.9, 1.1))
    parameters = get_test_parameter_set(2)
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_2d(sino, angles, **p)
        r.append(cutout(f))
    if WRITE_RES:
        write_results(myframe, r)
    assert np.allclose(np.array(r).flatten().view(float), get_results(myframe))


def test_2d_backprop_real():
    """
    Check if the real reconstruction matches the real part
    of the complex reconstruction.
    """
    sino, angles = create_test_sino_2d()
    parameters = get_test_parameter_set(2)
    # complex
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_2d(sino, angles, padval=0, **p)
        r.append(f)
    # real
    r2 = list()
    for p in parameters:
        f = odtbrain.backpropagate_2d(sino, angles, padval=0,
                                      onlyreal=True, **p)
        r2.append(f)
    assert np.allclose(np.array(r).real, np.array(r2))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import numpy as np


def compute_angle_weights_1d(angles: np.ndarray) -> np.ndarray:
    """Compute angular weights for tomographic reconstruction

    This method aims to address the issue of unevenly distributed angles
    in tomographic reconstruction. For algorithms, such as filtered
    backprojection, weighting each backprojection with a factor
    proportional to the angular distance to its neighbors dramatically
    improves the reconstructed image.

    Weights computation also takes into account these special cases:

    - Same angle present multiple times
    - Angular coverage larger than 180° (PI)

    Parameters
    ----------
    angles: 1d ndarray of length A
        Angles corresponding to the projections [rad]

    Returns
    -------
    weights: 1d ndarray of length A
        Weights for each angle; the mean of `weights` is one.

    Notes
    -----
    If one angle is passed multiple times `N` (e.g. `N=2`, 0° and 180°
    via `np.linspace(0, np.pi, 10, endpoint=True)`), then this angle will
    have a weight smaller than the other angles by a factor of `1/N`.

    This method is dupicated in the :ref:`radontea <radontea:index>`
    package. Even though, for ODT you normally need a coverage of 2 PI
    (instead of one PI in OPT), it makes sense here to wrap the coverage
    at PI. The idea is that opposite projections contribute similarly to
    the reconstruction and can be treated as PI-wrapped.
    """
    # copy and modulo np.pi
    # This is an array with values in [0, np.pi)
    angles = (angles.flatten() - np.min(angles)) % np.pi

    # If we have duplicate entries, we need to take them into account
    unq_angles, unq_reverse, unq_counts = np.unique(
        angles, return_inverse=True, return_counts=True)

    # sort the array
    srt_idx = np.argsort(unq_angles)
    srt_angles = unq_angles[srt_idx]
    srt_count = unq_counts[srt_idx]

    # compute weights for sorted angles
    da = (np.roll(srt_angles, -1) - np.roll(srt_angles, 1)) % np.pi
    sum_da = np.sum(da)

    # normalize with number of occurrences
    da /= srt_count
    srt_weights = da / sum_da * angles.size

    # Sort everything back where it belongs
    unq_weights = np.zeros_like(srt_weights)
    unq_weights[srt_idx] = srt_weights

    # Set the weights for each item in the original angles
    weights = unq_weights[unq_reverse]
    return weights



"""Translate reconstructed object functions to refractive index"""
import numpy as np


def odt_to_ri(f, res, nm):
    r"""Convert the ODT object function to refractive index

    In :abbr:`ODT (Optical Diffraction Tomography)`, the object function
    is defined by the Helmholtz equation

    .. math::

        f(\mathbf{r})  =  k_\mathrm{m}^2 \left[
            \left( \frac{n(\mathbf{r})}{n_\mathrm{m}} \right)^2 - 1
            \right]

    with :math:`k_\mathrm{m} = \frac{2\pi n_\mathrm{m}}{\lambda}`.
    By inverting this equation, we obtain the refractive index
    :math:`n(\mathbf{r})`.

    .. math::

        n(\mathbf{r})  = n_\mathrm{m}
            \sqrt{\frac{f(\mathbf{r})}{k_\mathrm{m}^2} + 1 }

    Parameters
    ----------
    f: n-dimensional ndarray
        The reconstructed object function :math:`f(\mathbf{r})`.
    res: float
        The size of the vacuum wave length :math:`\lambda` in pixels.
    nm: float
        The refractive index of the medium :math:`n_\mathrm{m}` that
        surrounds the object in :math:`f(\mathbf{r})`.

    Returns
    -------
    ri: n-dimensional ndarray
        The complex refractive index :math:`n(\mathbf{r})`.

    Notes
    -----
    Because this function computes the root of a complex number, there
    are several solutions to the refractive index. Always the positive
    (real) root of the refractive index is used.
    """
    km = (2 * np.pi * nm) / res
    ri = nm * np.sqrt(f / km**2 + 1)
    # Always take the positive root as the refractive index.
    # Because f can be imaginary, numpy cannot return the correct
    # positive root of f. However, we know that *ri* must be postive and
    # thus we take the absolute value of ri.
    # This also is what happens in Slaneys
    # diffract/Src/back.c in line 414.
    negrootcoord = np.where(ri.real < 0)
    ri[negrootcoord] *= -1
    return ri


def opt_to_ri(f, res, nm):
    r"""Convert the OPT object function to refractive index

    In :abbr:`OPT (Optical Projection Tomography)`, the object function
    is computed from the raw phase data. This method converts phase data
    to refractive index data.

    .. math::

        n(\mathbf{r})  = n_\mathrm{m} +
            \frac{f(\mathbf{r}) \cdot \lambda}{2 \pi}

    Parameters
    ----------
    f: n-dimensional ndarray
        The reconstructed object function :math:`f(\mathbf{r})`.
    res: float
        The size of the vacuum wave length :math:`\lambda` in pixels.
    nm: float
        The refractive index of the medium :math:`n_\mathrm{m}` that
        surrounds the object in :math:`f(\mathbf{r})`.

    Returns
    -------
    ri: n-dimensional ndarray
        The complex refractive index :math:`n(\mathbf{r})`.

    Notes
    -----
    This function is not meant to be used with diffraction tomography
    data. For ODT, use :py:func:`odt_to_ri` instead.
    """
    ri = nm + f / (2 * np.pi) * res
    return ri



"""Test 3D backpropagation with tilted axis of ration: matrices"""
import numpy as np

import odtbrain
import odtbrain._alg3d_bppt


def test_rotate_points_to_axis():
    # rotation of axis itself always goes to y-axis
    rot1 = odtbrain._alg3d_bppt.rotate_points_to_axis(
        points=[[1, 2, 3]], axis=[1, 2, 3])
    assert rot1[0][0] < 1e-14
    assert rot1[0][2] < 1e-14
    rot2 = odtbrain._alg3d_bppt.rotate_points_to_axis(
        points=[[-3, .6, .1]], axis=[-3, .6, .1])
    assert rot2[0][0] < 1e-14
    assert rot2[0][2] < 1e-14

    sq2 = np.sqrt(2)
    # rotation to 45deg about x
    points = [[0, 0, 1], [1, 0, 0], [1, 1, 0]]
    rot3 = odtbrain._alg3d_bppt.rotate_points_to_axis(
        points=points, axis=[0, 1, 1])
    assert np.allclose(rot3[0], [0, 1/sq2, 1/sq2])
    assert np.allclose(rot3[1], [1, 0, 0])
    assert np.allclose(rot3[2], [1, 1/sq2, -1/sq2])

    # rotation to 45deg about y
    points = [[0, 0, 1], [1, 0, 0], [0, -1, 0]]
    rot4 = odtbrain._alg3d_bppt.rotate_points_to_axis(
        points=points, axis=[1, 0, 1])
    assert np.allclose(rot4[0], [-.5, 1/sq2, .5])
    assert np.allclose(rot4[1], [.5, 1/sq2, -.5])
    assert np.allclose(rot4[2], [1/sq2, 0, 1/sq2])

    # Visualization
    # plt, Arrow3D = setup_mpl()
    # fig = plt.figure(figsize=(10,10))
    # ax = fig.add_subplot(111, projection='3d')
    # for vec in points:
    #     u,v,w = vec
    #     a = Arrow3D([0,u],[0,v],[0,w], mutation_scale=20,
    #                 lw=1, arrowstyle="-|>", color="k")
    #     ax.add_artist(a)
    # for vec in rot4:
    #     u,v,w = vec
    #     a = Arrow3D([0,u],[0,v],[0,w], mutation_scale=20, lw=1,
    #                 arrowstyle="-|>", color="b")
    #     ax.add_artist(a)
    #  radius=1
    #  ax.set_xlabel('X')
    #  ax.set_ylabel('Y')
    #  ax.set_zlabel('Z')
    # ax.set_xlim(-radius*1.5, radius*1.5)
    # ax.set_ylim(-radius*1.5, radius*1.5)
    # ax.set_zlim(-radius*1.5, radius*1.5)
    # plt.tight_layout()
    # plt.show()

    # rotation to -90deg about z
    points = [[0, 0, 1], [1, 0, 1], [1, -1, 0]]
    rot4 = odtbrain._alg3d_bppt.rotate_points_to_axis(
        points=points, axis=[1, 0, 0])
    assert np.allclose(rot4[0], [0, 0, 1])
    assert np.allclose(rot4[1], [0, 1, 1])
    assert np.allclose(rot4[2], [1, 1, 0])

    # negative axes
    # In this case, everything is rotated in the y-z plane
    # (this case is not physical for tomogrpahy)
    points = [[0, 0, 1], [1, 0, 0], [1, -1, 0]]
    rot4 = odtbrain._alg3d_bppt.rotate_points_to_axis(
        points=points, axis=[0, -1, 0])
    assert np.allclose(rot4[0], [0, 0, -1])
    assert np.allclose(rot4[1], [1, 0, 0])
    assert np.allclose(rot4[2], [1, 1, 0])


def test_rotation_matrix_from_point():
    """
    `rotation_matrix_from_point` generates a matrix that rotates a point at
    [0,0,1] to the position of the argument of the method.
    """
    sq2 = np.sqrt(2)
    # identity
    m1 = odtbrain._alg3d_bppt.rotation_matrix_from_point([0, 0, 1])
    assert np.allclose(np.dot(m1, [1, 2, 3]), [1, 2, 3])
    assert np.allclose(np.dot(m1, [-3, .5, -.6]), [-3, .5, -.6])

    # simple
    m2 = odtbrain._alg3d_bppt.rotation_matrix_from_point([0, 1, 1])
    assert np.allclose(np.dot(m2, [0, 0, 1]), [0, 1/sq2, 1/sq2])
    assert np.allclose(np.dot(m2, [0, 1, 1]), [0, sq2, 0])
    assert np.allclose(np.dot(m2, [1, 0, 0]), [1, 0, 0])

    # negative
    m3 = odtbrain._alg3d_bppt.rotation_matrix_from_point([-1, 1, 0])
    assert np.allclose(np.dot(m3, [1, 0, 0]), [0, 0, -1])
    assert np.allclose(np.dot(m3, [0, 1, 1]), [0, sq2, 0])
    assert np.allclose(np.dot(m3, [0, -1/sq2, -1/sq2]), [0, -1, 0])
    assert np.allclose(np.dot(m3, [0, 1/sq2, -1/sq2]), [-1, 0, 0])
    assert np.allclose(np.dot(m3, [0, -1/sq2, 1/sq2]), [1, 0, 0])


def setup_mpl():
    import matplotlib.pylab as plt
    from mpl_toolkits.mplot3d import Axes3D  # noqa F01
    from matplotlib.patches import FancyArrowPatch
    from mpl_toolkits.mplot3d import proj3d

    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
            self._verts3d = xs, ys, zs

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, _zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            FancyArrowPatch.draw(self, renderer)

    return plt, Arrow3D


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()

    import scipy.ndimage
    plt, Arrow3D = setup_mpl()

    # Testarray
    N = 50
    A = 41
    proj = np.zeros((N, N, N))
    proj[(N)/2, (N)/2, :(N)/2] = np.abs(np.linspace(-10, 1, (N)/2))

    # By default, the rotational axis in _Back_3D_tilted is the y-axis.
    # Define a rotational axis with a slight offset in x and in z.
    axis = np.array([0.0, 1, .1])
    axis /= np.sqrt(np.sum(axis**2))

    # Now, obtain the 3D angles that are equally distributed on the unit
    # sphere and correspond to the positions of projections that we would
    # measure.
    angles = np.linspace(0, 2*np.pi, A, endpoint=False)
    # The first point in that array will be in the x-z-plane.
    points = odtbrain._alg3d_bppt.sphere_points_from_angles_and_tilt(
        angles, axis)

    # The following steps are exactly those that are used in
    # odtbrain._alg3d_bppt.backpropagate_3d_tilted
    # to perform 3D reconstruction with tilted angles.
    u, v, w = axis
    theta = np.arccos(v)

    # We need three rotations.

    # IMPROTANT:
    # We perform the reconstruction such that the rotational axis
    # is equal to the y-axis! This is easier than implementing a
    # rotation about the rotational axis and tilting with theta
    # before and afterwards.

    # This is the rotation that tilts the projection in the
    # direction of the rotation axis (new y-axis).
    Rtilt = np.array([
        [1,             0,              0],
        [0, np.cos(theta),  np.sin(theta)],
        [0, -np.sin(theta),  np.cos(theta)],
    ])

    out = np.zeros((N, N, N))

    vectors = []

    for ang, pnt in zip(angles, points):
        Rcircle = np.array([
            [np.cos(ang),  0, np.sin(ang)],
            [0, 1,           0],
            [-np.sin(ang), 0, np.cos(ang)],
        ])

        DR = np.dot(Rtilt, Rcircle)

        # pnt are already rotated by R1
        vectors.append(np.dot(Rtilt, pnt))

        # We need to give this rotation the correct offset
        c = 0.5*np.array(proj.shape)
        offset = c-c.dot(DR.T)
        rotate = scipy.ndimage.interpolation.affine_transform(
            proj, DR, offset=offset,
            mode="constant", cval=0, order=2)
        proj *= 0.98
        out += rotate

    # visualize the axes
    out[0, 0, 0] = np.max(out)  # origin
    out[-1, 0, 0] = np.max(out)/2  # x
    out[0, -1, 0] = np.max(out)/3  # z
    # show arrows pointing at projection directions
    # (should form cone aligned with y)
    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(111, projection='3d')
    for vec in vectors:
        u, v, w = vec
        a = Arrow3D([0, u], [0, v], [0, w],
                    mutation_scale=20, lw=1, arrowstyle="-|>")
        ax.add_artist(a)

    radius = 1
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.set_xlim(-radius*1.5, radius*1.5)
    ax.set_ylim(-radius*1.5, radius*1.5)
    ax.set_zlim(-radius*1.5, radius*1.5)
    plt.tight_layout()
    plt.show()



"""HL60 cell

The quantitative phase data of an HL60 S/4 cell were recorded using
:abbr:`QLSI (quadri-wave lateral shearing interferometry)`.
The original dataset was used in a previous publication
:cite:`Schuermann2017` to illustrate the capabilities of combined
fluorescence and refractive index tomography.

The example data set is already aligned and background-corrected as
described in the original publication and the fluorescence data are
not included. The lzma-archive contains the sinogram data stored in
the :ref:`qpimage <qpimage:index>` file format and the rotational
positions of each sinogram image as a text file.

The figure reproduces parts of figure 4 of the original manuscript.
Note that minor deviations from the original figure can be attributed
to the strong compression (scale offset filter) and due to the fact
that the original sinogram images were cropped from 196x196 px to
140x140 px (which in particular affects the background-part of the
refractive index histogram).

The raw data is available
`on figshare <https://doi.org/10.6084/m9.figshare.8055407.v1>`
(hl60_sinogram_qpi.h5).
"""
import pathlib
import tarfile
import tempfile

import matplotlib.pylab as plt
import numpy as np
import odtbrain as odt
import qpimage

from example_helper import get_file, extract_lzma


if __name__ == "__main__":
    # ascertain the data
    path = get_file("qlsi_3d_hl60-cell_A140.tar.lzma")
    tarf = extract_lzma(path)
    tdir = tempfile.mkdtemp(prefix="odtbrain_example_")

    with tarfile.open(tarf) as tf:
        tf.extract("series.h5", path=tdir)
        angles = np.loadtxt(tf.extractfile("angles.txt"))

    # extract the complex field sinogram from the qpimage series data
    h5file = pathlib.Path(tdir) / "series.h5"
    with qpimage.QPSeries(h5file=h5file, h5mode="r") as qps:
        qp0 = qps[0]
        meta = qp0.meta
        sino = np.zeros((len(qps), qp0.shape[0], qp0.shape[1]),
                        dtype=np.complex)
        for ii in range(len(qps)):
            sino[ii] = qps[ii].field

    # perform backpropagation
    u_sinR = odt.sinogram_as_rytov(sino)
    res = meta["wavelength"] / meta["pixel size"]
    nm = meta["medium index"]

    fR = odt.backpropagate_3d(uSin=u_sinR,
                              angles=angles,
                              res=res,
                              nm=nm)

    ri = odt.odt_to_ri(fR, res, nm)

    # plot results
    ext = meta["pixel size"] * 1e6 * 70
    kw = {"vmin": ri.real.min(),
          "vmax": ri.real.max(),
          "extent": [-ext, ext, -ext, ext]}
    fig, axes = plt.subplots(1, 3, figsize=(8, 2.5))
    axes[0].imshow(ri[70, :, :].real, **kw)
    axes[0].set_xlabel("x [µm]")
    axes[0].set_ylabel("y [µm]")

    x = np.linspace(-ext, ext, 140)
    axes[1].plot(x, ri[70, :, 70], label="line plot x=0")
    axes[1].plot(x, ri[70, 70, :], label="line plot y=0")
    axes[1].set_xlabel("distance from center [µm]")
    axes[1].set_ylabel("refractive index")
    axes[1].legend()

    hist, xh = np.histogram(ri.real, bins=100)
    axes[2].plot(xh[1:], hist)
    axes[2].set_yscale('log')
    axes[2].set_xlabel("refractive index")
    axes[2].set_ylabel("histogram")

    plt.tight_layout()
    plt.show()



"""Test Fourier mapping algorithm"""
import sys

import numpy as np
import odtbrain
import pytest

from common_methods import create_test_sino_2d, cutout, \
    get_test_parameter_set, write_results, get_results

WRITE_RES = False


@pytest.mark.xfail(True, reason="Unexplained issue #13 and new algorithm #19")
@pytest.mark.filterwarnings(
    "ignore::odtbrain.warn.ImplementationAmbiguousWarning")
def test_2d_fmap():
    myframe = sys._getframe()
    sino, angles = create_test_sino_2d()
    parameters = get_test_parameter_set(1)
    r = []
    for p in parameters:
        f = odtbrain.fourier_map_2d(sino, angles, **p)
        r.append(cutout(f))
    if WRITE_RES:
        write_results(myframe, r)
    assert np.allclose(np.array(r).flatten().view(float), get_results(myframe))



"""FDTD cell phantom

The *in silico* data set was created with the
:abbr:`FDTD (Finite Difference Time Domain)` software `meep`_. The data
are 1D projections of a 2D refractive index phantom. The
reconstruction of the refractive index with the Rytov approximation
is in good agreement with the phantom that was used in the
simulation.

.. note::

   In the initial version of this example, I did not notice that the
   1D fields in the data file were actually not in-focus. Back then,
   I used an auto-focusing algorithm to numerically focus the sinogram
   data to the rotation center, and this distance was inaccurate.
   Upon closer inspection of the reconstructed image, it turned out that
   the focus position is still about 0.5 wavelengths away from the
   rotation center. I have added this information to the `fdtd_info.txt`
   file in the data archive.

   The take-home message is that for a proper tomographic reconstruction
   it is important that the distance between the rotation center and
   the focus of the sinogram is known. If this is not the case, it might
   make sense to sweep the reconstruction distance and use an
   appropriate metric to determine the best reconstruction.

.. _`meep`: http://ab-initio.mit.edu/wiki/index.php/Meep
"""
import matplotlib.pylab as plt
import numpy as np
import odtbrain as odt

from example_helper import load_data


sino, angles, phantom, cfg = load_data("fdtd_2d_sino_A100_R13.zip",
                                       f_angles="fdtd_angles.txt",
                                       f_sino_imag="fdtd_imag.txt",
                                       f_sino_real="fdtd_real.txt",
                                       f_info="fdtd_info.txt",
                                       f_phantom="fdtd_phantom.txt",
                                       )

print("Example: Backpropagation from 2D FDTD simulations")
print("Refractive index of medium:", cfg["nm"])
print("Measurement position from object center in wavelengths:", cfg["lD"])
print("Wavelength sampling:", cfg["res"])
print("Performing backpropagation.")

# Apply the Rytov approximation
sino_rytov = odt.sinogram_as_rytov(sino)

# perform backpropagation to obtain object function f
f = odt.backpropagate_2d(uSin=sino_rytov,
                         angles=angles,
                         res=cfg["res"],
                         nm=cfg["nm"],
                         lD=cfg["lD"] * cfg["res"]
                         )

# compute refractive index n from object function
n = odt.odt_to_ri(f, res=cfg["res"], nm=cfg["nm"])

# compare phantom and reconstruction in plot
fig, axes = plt.subplots(1, 3, figsize=(8, 2.8))

axes[0].set_title("FDTD phantom")
axes[0].imshow(phantom, vmin=phantom.min(), vmax=phantom.max())
sino_phase = np.unwrap(np.angle(sino), axis=1)

axes[1].set_title("phase sinogram")
axes[1].imshow(sino_phase, vmin=sino_phase.min(), vmax=sino_phase.max(),
               aspect=sino.shape[1] / sino.shape[0],
               cmap="coolwarm")
axes[1].set_xlabel("detector")
axes[1].set_ylabel("angle [rad]")

axes[2].set_title("reconstructed image")
axes[2].imshow(n.real, vmin=phantom.min(), vmax=phantom.max())

# set y ticks for sinogram
labels = np.linspace(0, 2 * np.pi, len(axes[1].get_yticks()))
labels = ["{:.2f}".format(i) for i in labels]
axes[1].set_yticks(np.linspace(0, len(angles), len(labels)))
axes[1].set_yticklabels(labels)

plt.tight_layout()
plt.show()



# flake8: noqa: F401
"""Algorithms for scalar diffraction tomography"""
from ._alg2d_bpp import backpropagate_2d
from ._alg2d_fmp import fourier_map_2d
from ._alg2d_int import integrate_2d

from ._alg3d_bpp import backpropagate_3d
from ._alg3d_bppt import backpropagate_3d_tilted

from ._prepare_sino import sinogram_as_radon, sinogram_as_rytov
from ._translate_ri import odt_to_ri, opt_to_ri
from ._version import version as __version__

from . import apple
from . import util
from . import warn


__author__ = "Paul Müller"
__license__ = "BSD (3 clause)"



"""Test save memory options"""
import numpy as np

import odtbrain

from common_methods import create_test_sino_3d, get_test_parameter_set


def test_back3d():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    parameters = get_test_parameter_set(2)
    # complex
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                      dtype=float,
                                      save_memory=False, **p)
        r.append(f)
    # real
    r2 = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                      dtype=float,
                                      save_memory=True, **p)
        r2.append(f)
    assert np.allclose(np.array(r), np.array(r2))


def test_back3d_tilted():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    parameters = get_test_parameter_set(2)
    # complex
    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d_tilted(sino, angles, padval=0,
                                             dtype=float,
                                             save_memory=False, **p)
        r.append(f)
    # real
    r2 = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d_tilted(sino, angles, padval=0,
                                             dtype=float,
                                             save_memory=True, **p)
        r2.append(f)

    assert np.allclose(np.array(r), np.array(r2))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



r"""Mie sphere
The *in silico* data set was created with the Mie calculation software
`GMM-field`_. The data consist of a two-dimensional projection of a
sphere with radius :math:`R=14\lambda`,
refractive index :math:`n_\mathrm{sph}=1.006`,
embedded in a medium of refractive index :math:`n_\mathrm{med}=1.0`
onto a detector which is :math:`l_\mathrm{D} = 20\lambda` away from the
center of the sphere.

The package :mod:`nrefocus` must be used to numerically focus
the detected field prior to the 3D backpropagation with ODTbrain.
In :func:`odtbrain.backpropagate_3d`, the parameter `lD` must
be set to zero (:math:`l_\mathrm{D}=0`).

The figure shows the 3D reconstruction from Mie simulations of a
perfect sphere using 200 projections. Missing angle artifacts are
visible along the :math:`y`-axis due to the :math:`2\pi`-only
coverage in 3D Fourier space.

.. _`GMM-field`: https://code.google.com/p/scatterlib/wiki/Nearfield
"""
import matplotlib.pylab as plt
import nrefocus
import numpy as np

import odtbrain as odt

from example_helper import load_data


if __name__ == "__main__":
    Ex, cfg = load_data("mie_3d_sphere_field.zip",
                        f_sino_imag="mie_sphere_imag.txt",
                        f_sino_real="mie_sphere_real.txt",
                        f_info="mie_info.txt")

    # Manually set number of angles:
    A = 200

    print("Example: Backpropagation from 3D Mie scattering")
    print("Refractive index of medium:", cfg["nm"])
    print("Measurement position from object center:", cfg["lD"])
    print("Wavelength sampling:", cfg["res"])
    print("Number of angles for reconstruction:", A)
    print("Performing backpropagation.")

    # Reconstruction angles
    angles = np.linspace(0, 2 * np.pi, A, endpoint=False)

    # Perform focusing
    Ex = nrefocus.refocus(Ex,
                          d=-cfg["lD"]*cfg["res"],
                          nm=cfg["nm"],
                          res=cfg["res"],
                          )

    # Create sinogram
    u_sin = np.tile(Ex.flat, A).reshape(A, int(cfg["size"]), int(cfg["size"]))

    # Apply the Rytov approximation
    u_sinR = odt.sinogram_as_rytov(u_sin)

    # Backpropagation
    fR = odt.backpropagate_3d(uSin=u_sinR,
                              angles=angles,
                              res=cfg["res"],
                              nm=cfg["nm"],
                              lD=0,
                              padfac=2.1,
                              save_memory=True)

    # RI computation
    nR = odt.odt_to_ri(fR, cfg["res"], cfg["nm"])

    # Plotting
    fig, axes = plt.subplots(2, 3, figsize=(8, 5))
    axes = np.array(axes).flatten()
    # field
    axes[0].set_title("Mie field phase")
    axes[0].set_xlabel("detector x")
    axes[0].set_ylabel("detector y")
    axes[0].imshow(np.angle(Ex), cmap="coolwarm")
    axes[1].set_title("Mie field amplitude")
    axes[1].set_xlabel("detector x")
    axes[1].set_ylabel("detector y")
    axes[1].imshow(np.abs(Ex), cmap="gray")

    # line plot
    axes[2].set_title("line plots")
    axes[2].set_xlabel("distance [px]")
    axes[2].set_ylabel("real refractive index")
    center = int(cfg["size"] / 2)
    x = np.arange(cfg["size"]) - center
    axes[2].plot(x, nR[:, center, center].real, label="x")
    axes[2].plot(x, nR[center, center, :].real, label="z")
    axes[2].plot(x, nR[center, :, center].real, label="y")
    axes[2].legend(loc=4)
    axes[2].set_xlim((-center, center))
    dn = abs(cfg["nsph"] - cfg["nm"])
    axes[2].set_ylim((cfg["nm"] - dn / 10, cfg["nsph"] + dn))
    axes[2].ticklabel_format(useOffset=False)

    # cross sections
    axes[3].set_title("RI reconstruction\nsection at x=0")
    axes[3].set_xlabel("z")
    axes[3].set_ylabel("y")
    axes[3].imshow(nR[center, :, :].real)

    axes[4].set_title("RI reconstruction\nsection at y=0")
    axes[4].set_xlabel("x")
    axes[4].set_ylabel("z")
    axes[4].imshow(nR[:, center, :].real)

    axes[5].set_title("RI reconstruction\nsection at z=0")
    axes[5].set_xlabel("y")
    axes[5].set_ylabel("x")
    axes[5].imshow(nR[:, :, center].real)

    plt.tight_layout()
    plt.show()



"""Mie off-center cylinder

The *in silico* data set was created with the
softare `miefield  <https://github.com/RI-imaging/miefield>`_.
The data are 1D projections of an off-center cylinder of constant
refractive index. The Born approximation is error-prone due to
a relatively large radius of the cylinder (30 wavelengths) and
a refractive index difference of 0.006 between cylinder and
surrounding medium. The reconstruction of the refractive index
with the Rytov approximation is in good agreement with the
input data. When only 50 projections are used for the reconstruction,
artifacts appear. These vanish when more projections are used for
the reconstruction.
"""
import matplotlib.pylab as plt
import numpy as np

import odtbrain as odt

from example_helper import load_data


# simulation data
sino, angles, cfg = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
                              f_sino_imag="sino_imag.txt",
                              f_sino_real="sino_real.txt",
                              f_angles="mie_angles.txt",
                              f_info="mie_info.txt")
A, size = sino.shape

# background sinogram computed with Mie theory
# miefield.GetSinogramCylinderRotation(radius, nmed, nmed, lD, lC, size, A,res)
u0 = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
               f_sino_imag="u0_imag.txt",
               f_sino_real="u0_real.txt")
# create 2d array
u0 = np.tile(u0, size).reshape(A, size).transpose()

# background field necessary to compute initial born field
# u0_single = mie.GetFieldCylinder(radius, nmed, nmed, lD, size, res)
u0_single = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
                      f_sino_imag="u0_single_imag.txt",
                      f_sino_real="u0_single_real.txt")

print("Example: Backpropagation from 2D Mie simulations")
print("Refractive index of medium:", cfg["nmed"])
print("Measurement position from object center:", cfg["lD"])
print("Wavelength sampling:", cfg["res"])
print("Performing backpropagation.")

# Set measurement parameters
# Compute scattered field from cylinder
radius = cfg["radius"]  # wavelengths
nmed = cfg["nmed"]
ncyl = cfg["ncyl"]

lD = cfg["lD"]  # measurement distance in wavelengths
lC = cfg["lC"]  # displacement from center of image
size = cfg["size"]
res = cfg["res"]  # px/wavelengths
A = cfg["A"]  # number of projections

x = np.arange(size) - size / 2
X, Y = np.meshgrid(x, x)
rad_px = radius * res
phantom = np.array(((Y - lC * res)**2 + X**2) < rad_px**2,
                   dtype=float) * (ncyl - nmed) + nmed

# Born
u_sinB = (sino / u0 * u0_single - u0_single)  # fake born
fB = odt.backpropagate_2d(u_sinB, angles, res, nmed, lD * res)
nB = odt.odt_to_ri(fB, res, nmed)

# Rytov
u_sinR = odt.sinogram_as_rytov(sino / u0)
fR = odt.backpropagate_2d(u_sinR, angles, res, nmed, lD * res)
nR = odt.odt_to_ri(fR, res, nmed)

# Rytov 50
u_sinR50 = odt.sinogram_as_rytov((sino / u0)[::5, :])
fR50 = odt.backpropagate_2d(u_sinR50, angles[::5], res, nmed, lD * res)
nR50 = odt.odt_to_ri(fR50, res, nmed)

# Plot sinogram phase and amplitude
ph = odt.sinogram_as_radon(sino / u0)

am = np.abs(sino / u0)

# prepare plot
vmin = np.min(np.array([phantom, nB.real, nR50.real, nR.real]))
vmax = np.max(np.array([phantom, nB.real, nR50.real, nR.real]))

fig, axes = plt.subplots(2, 3, figsize=(8, 5))
axes = np.array(axes).flatten()

phantommap = axes[0].imshow(phantom, vmin=vmin, vmax=vmax)
axes[0].set_title("phantom \n(non-centered cylinder)")

amplmap = axes[1].imshow(am, cmap="gray")
axes[1].set_title("amplitude sinogram \n(background-corrected)")

phasemap = axes[2].imshow(ph, cmap="coolwarm")
axes[2].set_title("phase sinogram [rad] \n(background-corrected)")

axes[3].imshow(nB.real, vmin=vmin, vmax=vmax)
axes[3].set_title("reconstruction (Born) \n(250 projections)")

axes[4].imshow(nR50.real, vmin=vmin, vmax=vmax)
axes[4].set_title("reconstruction (Rytov) \n(50 projections)")

axes[5].imshow(nR.real, vmin=vmin, vmax=vmax)
axes[5].set_title("reconstruction (Rytov) \n(250 projections)")

# color bars
cbkwargs = {"fraction": 0.045}
plt.colorbar(phantommap, ax=axes[0], **cbkwargs)
plt.colorbar(amplmap, ax=axes[1], **cbkwargs)
plt.colorbar(phasemap, ax=axes[2], **cbkwargs)
plt.colorbar(phantommap, ax=axes[3], **cbkwargs)
plt.colorbar(phantommap, ax=axes[4], **cbkwargs)
plt.colorbar(phantommap, ax=axes[5], **cbkwargs)

plt.tight_layout()
plt.show()



"""Test apple core correction"""
import multiprocessing as mp
import sys

import numpy as np

import odtbrain

from common_methods import create_test_sino_3d, cutout, \
    get_test_parameter_set, write_results, get_results


WRITE_RES = False


def test_apple_core_3d_values():
    try:
        odtbrain.apple.apple_core_3d(shape=(10, 10, 5),
                                     res=.1,
                                     nm=1)
    except ValueError:
        pass
    else:
        assert False, "bad input shape should raise ValueError"


def test_correct_counter():
    count = mp.Value("I", lock=True)
    max_count = mp.Value("I", lock=True)

    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                  dtype=float,
                                  copy=False, **p)
    odtbrain.apple.correct(f=f,
                           res=p["res"],
                           nm=p["nm"],
                           enforce_envelope=.95,
                           max_iter=100,
                           min_diff=0.01,
                           count=count,
                           max_count=max_count)

    assert count.value == max_count.value


def test_correct_reproduce():
    myframe = sys._getframe()
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    sryt = odtbrain.sinogram_as_rytov(uSin=sino, u0=1, align=False)
    f = odtbrain.backpropagate_3d(sryt, angles, padval=0,
                                  dtype=float,
                                  copy=False, **p)
    fc = odtbrain.apple.correct(f=f,
                                res=p["res"],
                                nm=p["nm"],
                                enforce_envelope=.95,
                                max_iter=100,
                                min_diff=0.01)
    fo = cutout(fc)
    fo = np.array(fo, dtype=np.complex128)

    if WRITE_RES:
        write_results(myframe, fo)

    data = fo.flatten().view(float)
    assert np.allclose(data, get_results(myframe), atol=2.E-6)


def test_correct_values():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    f = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                  dtype=float,
                                  copy=False, **p)
    try:
        odtbrain.apple.correct(f=f,
                               res=p["res"],
                               nm=p["nm"],
                               enforce_envelope=1.05,
                               )
    except ValueError:
        pass
    else:
        assert False, "`enforce_envelope` must be in [0, 1]"


def test_envelope_gauss_shape():
    """Make sure non-cubic input shape works"""
    # non-cubic reconstruction volume (1st and 3rd axis still have same length)
    shape = (60, 50, 60)
    ftdata = np.ones(shape)
    core = odtbrain.apple.apple_core_3d(shape=shape, res=.1, nm=1)
    envlp = odtbrain.apple.envelope_gauss(ftdata=ftdata, core=core)
    assert envlp.shape == shape


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test slow integration algorithm"""
import sys

import numpy as np
import odtbrain

import pytest

from common_methods import create_test_sino_2d, cutout, \
    get_test_parameter_set, write_results, get_results

WRITE_RES = False


@pytest.mark.filterwarnings(
    "ignore::odtbrain.warn.DataUndersampledWarning")
def test_2d_integrate():
    myframe = sys._getframe()
    sino, angles = create_test_sino_2d()
    parameters = get_test_parameter_set(2)
    r = list()

    for p in parameters:
        f = odtbrain.integrate_2d(sino, angles, **p)
        r.append(cutout(f))

    if WRITE_RES:
        write_results(myframe, r)
    assert np.allclose(np.array(r).flatten().view(float), get_results(myframe))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Tests refractive index conversion techniques"""
import sys

import numpy as np

import odtbrain
from odtbrain._prepare_sino import divmod_neg

from common_methods import write_results, get_results

WRITE_RES = False


def get_test_data_set():
    """returns 3D array and parameters"""
    ln = 10
    f = np.arange(ln**3).reshape(ln, ln, ln)
    f = f + np.linspace(1, 2, ln)
    res = 7
    nm = 1.34
    return f, res, nm


def get_test_data_set_sino(rytov=False):
    """returns 3D array"""
    ln = 10
    a = 2
    sino = np.arange(ln*ln*a).reshape(a, ln, ln) / (ln*ln*a) * \
        np.exp(1j*np.arange(ln*ln*a).reshape(a, ln, ln))
    if rytov:
        sino[0, 0, 0] = .1
    return sino


def negative_modulo_rest(a, b):
    """returns modulo with closest result to zero"""
    q = np.array(a / b, dtype=int)
    r = a - b * q

    # make sure r is close to zero
    wrong = np.where(np.abs(r) > b/2)
    r[wrong] -= b * np.sign(r[wrong])
    return r


def negative_modulo_rest_imag(x, b):
    """only modulo the imaginary part"""
    a = x.imag
    return x.real + 1j*negative_modulo_rest(a, b)


def test_odt_to_ri():
    myframe = sys._getframe()
    f, res, nm = get_test_data_set()
    ri = odtbrain.odt_to_ri(f=f, res=res, nm=nm)
    if WRITE_RES:
        write_results(myframe, ri)
    assert np.allclose(np.array(ri).flatten().view(
        float), get_results(myframe))
    # Also test 2D version
    ri2d = odtbrain.odt_to_ri(f=f[0], res=res, nm=nm)
    assert np.allclose(ri2d, ri[0])


def test_opt_to_ri():
    myframe = sys._getframe()
    f, res, nm = get_test_data_set()
    ri = odtbrain.opt_to_ri(f=f, res=res, nm=nm)
    if WRITE_RES:
        write_results(myframe, ri)
    assert np.allclose(np.array(ri).flatten().view(
        float), get_results(myframe))
    # Also test 2D version
    ri2d = odtbrain.opt_to_ri(f=f[0], res=res, nm=nm)
    assert np.allclose(ri2d, ri[0])


def test_sino_radon():
    myframe = sys._getframe()
    sino = get_test_data_set_sino()
    rad = odtbrain.sinogram_as_radon(sino)
    twopi = 2*np.pi
    # When moving from unwrap to skimage, there was an offset introduced.
    # Since this particular array is not flat at the borders, there is no
    # correct way here. We just subtract 2PI.
    # 2019-04-18: It turns out that on Windows, this is not the case.
    # Hence, we only subtract 2PI if the minimum of the array is above
    # 2PI..
    if rad.min() > twopi:
        rad -= twopi
    if WRITE_RES:
        write_results(myframe, rad)
    assert np.allclose(np.array(rad).flatten().view(
        float), get_results(myframe))
    # Check the 3D result with the 2D result. They should be the same except
    # for a multiple of 2PI offset, because odtbrain._align_unwrapped
    # subtracts the background such that the minimum phase change is closest
    # to zero.
    # 2D A
    rad2d = odtbrain.sinogram_as_radon(sino[:, :, 0])
    assert np.allclose(0, negative_modulo_rest(
        rad2d - rad[:, :, 0], twopi), atol=1e-6)
    # 2D B
    rad2d2 = odtbrain.sinogram_as_radon(sino[:, 0, :])
    assert np.allclose(0, negative_modulo_rest(
        rad2d2 - rad[:, 0, :], twopi), atol=1e-6)


def test_sino_rytov():
    myframe = sys._getframe()
    sino = get_test_data_set_sino(rytov=True)
    ryt = odtbrain.sinogram_as_rytov(sino)
    twopi = 2*np.pi
    if WRITE_RES:
        write_results(myframe, ryt)
    # When moving from unwrap to skimage, there was an offset introduced.
    # Since this particular array is not flat at the borders, there is no
    # correct way here. We just subtract 2PI.
    # 2019-04-18: It turns out that on Windows, this is not the case.
    # Hence, we only subtract 2PI if the minimum of the array is above
    # 2PI..
    if ryt.imag.min() > twopi:
        ryt.imag -= twopi
    assert np.allclose(np.array(ryt).flatten().view(
        float), get_results(myframe))
    # Check the 3D result with the 2D result. They should be the same except
    # for a multiple of 2PI offset, because odtbrain._align_unwrapped
    # subtracts the background such that the median phase change is closest
    # to zero.
    # 2D A
    ryt2d = odtbrain.sinogram_as_rytov(sino[:, :, 0])
    assert np.allclose(0, negative_modulo_rest_imag(
        ryt2d - ryt[:, :, 0], twopi).view(float), atol=1e-6)
    # 2D B
    ryt2d2 = odtbrain.sinogram_as_rytov(sino[:, 0, :])
    assert np.allclose(0, negative_modulo_rest_imag(
        ryt2d2 - ryt[:, 0, :], twopi).view(float), atol=1e-6)


def test_divmod_neg():
    assert np.allclose(divmod_neg(0, 2*np.pi), (0, 0))
    assert np.allclose(divmod_neg(-1e-17, 2*np.pi), (0, 0))
    assert np.allclose(divmod_neg(1e-17, 2*np.pi), (0, 0))
    assert np.allclose(divmod_neg(-.1, 2*np.pi), (0, -.1))
    assert np.allclose(divmod_neg(.1, 2*np.pi), (0, .1))
    assert np.allclose(divmod_neg(3*np.pi, 2*np.pi), (1, np.pi))
    assert np.allclose(divmod_neg(-.99*np.pi, 2*np.pi), (0, -.99*np.pi))
    assert np.allclose(divmod_neg(-1.01*np.pi, 2*np.pi), (-1, .99*np.pi))


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""FDTD cell phantom
The *in silico* data set was created with the
:abbr:`FDTD (Finite Difference Time Domain)` software `meep`_. The data
are 2D projections of a 3D refractive index phantom. The reconstruction
of the refractive index with the Rytov approximation is in good
agreement with the phantom that was used in the simulation. The data
are downsampled by a factor of two. The rotational axis is the `y`-axis.
A total of 180 projections are used for the reconstruction. A detailed
description of this phantom is given in :cite:`Mueller2015`.

.. _`meep`: http://ab-initio.mit.edu/wiki/index.php/Meep
"""
import matplotlib.pylab as plt
import numpy as np

import odtbrain as odt

from example_helper import load_data


if __name__ == "__main__":
    sino, angles, phantom, cfg = \
        load_data("fdtd_3d_sino_A180_R6.500.tar.lzma")

    A = angles.shape[0]

    print("Example: Backpropagation from 3D FDTD simulations")
    print("Refractive index of medium:", cfg["nm"])
    print("Measurement position from object center:", cfg["lD"])
    print("Wavelength sampling:", cfg["res"])
    print("Number of projections:", A)
    print("Performing backpropagation.")

    # Apply the Rytov approximation
    sinoRytov = odt.sinogram_as_rytov(sino)

    # perform backpropagation to obtain object function f
    f = odt.backpropagate_3d(uSin=sinoRytov,
                             angles=angles,
                             res=cfg["res"],
                             nm=cfg["nm"],
                             lD=cfg["lD"]
                             )

    # compute refractive index n from object function
    n = odt.odt_to_ri(f, res=cfg["res"], nm=cfg["nm"])

    sx, sy, sz = n.shape
    px, py, pz = phantom.shape

    sino_phase = np.angle(sino)

    # compare phantom and reconstruction in plot
    fig, axes = plt.subplots(2, 3, figsize=(8, 4))
    kwri = {"vmin": n.real.min(), "vmax": n.real.max()}
    kwph = {"vmin": sino_phase.min(), "vmax": sino_phase.max(),
            "cmap": "coolwarm"}

    # Phantom
    axes[0, 0].set_title("FDTD phantom center")
    rimap = axes[0, 0].imshow(phantom[px // 2], **kwri)
    axes[0, 0].set_xlabel("x")
    axes[0, 0].set_ylabel("y")

    axes[1, 0].set_title("FDTD phantom nucleolus")
    axes[1, 0].imshow(phantom[int(px / 2 + 2 * cfg["res"])], **kwri)
    axes[1, 0].set_xlabel("x")
    axes[1, 0].set_ylabel("y")

    # Sinogram
    axes[0, 1].set_title("phase projection")
    phmap = axes[0, 1].imshow(sino_phase[A // 2, :, :], **kwph)
    axes[0, 1].set_xlabel("detector x")
    axes[0, 1].set_ylabel("detector y")

    axes[1, 1].set_title("sinogram slice")
    axes[1, 1].imshow(sino_phase[:, :, sino.shape[2] // 2],
                      aspect=sino.shape[1] / sino.shape[0], **kwph)
    axes[1, 1].set_xlabel("detector y")
    axes[1, 1].set_ylabel("angle [rad]")
    # set y ticks for sinogram
    labels = np.linspace(0, 2 * np.pi, len(axes[1, 1].get_yticks()))
    labels = ["{:.2f}".format(i) for i in labels]
    axes[1, 1].set_yticks(np.linspace(0, len(angles), len(labels)))
    axes[1, 1].set_yticklabels(labels)

    axes[0, 2].set_title("reconstruction center")
    axes[0, 2].imshow(n[sx // 2].real, **kwri)
    axes[0, 2].set_xlabel("x")
    axes[0, 2].set_ylabel("y")

    axes[1, 2].set_title("reconstruction nucleolus")
    axes[1, 2].imshow(n[int(sx / 2 + 2 * cfg["res"])].real, **kwri)
    axes[1, 2].set_xlabel("x")
    axes[1, 2].set_ylabel("y")

    # color bars
    cbkwargs = {"fraction": 0.045,
                "format": "%.3f"}
    plt.colorbar(phmap, ax=axes[0, 1], **cbkwargs)
    plt.colorbar(phmap, ax=axes[1, 1], **cbkwargs)
    plt.colorbar(rimap, ax=axes[0, 0], **cbkwargs)
    plt.colorbar(rimap, ax=axes[1, 0], **cbkwargs)
    plt.colorbar(rimap, ax=axes[0, 2], **cbkwargs)
    plt.colorbar(rimap, ax=axes[1, 2], **cbkwargs)

    plt.tight_layout()
    plt.show()



"""Mie cylinder with unevenly spaced angles

Angular weighting can significantly improve reconstruction quality
when the angular projections are sampled at non-equidistant
intervals :cite:`Tam1981`. The *in silico* data set was created with
the softare `miefield  <https://github.com/RI-imaging/miefield>`_.
The data are 1D projections of a non-centered cylinder of constant
refractive index 1.339 embedded in water with refractive index 1.333.
The first column shows the used sinograms (missing angles are displayed
as zeros) that were created from the original sinogram with 250
projections. The second column shows the reconstruction without angular
weights and the third column shows the reconstruction with angular
weights. The keyword argument `weight_angles` was introduced in version
0.1.1.
"""
import matplotlib.pylab as plt
import numpy as np
from skimage.restoration import unwrap_phase

import odtbrain as odt

from example_helper import load_data


sino, angles, cfg = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
                              f_angles="mie_angles.txt",
                              f_sino_real="sino_real.txt",
                              f_sino_imag="sino_imag.txt",
                              f_info="mie_info.txt")
A, size = sino.shape

# background sinogram computed with Mie theory
# miefield.GetSinogramCylinderRotation(radius, nmed, nmed, lD, lC, size, A,res)
u0 = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
               f_sino_imag="u0_imag.txt",
               f_sino_real="u0_real.txt")
# create 2d array
u0 = np.tile(u0, size).reshape(A, size).transpose()

# background field necessary to compute initial born field
# u0_single = mie.GetFieldCylinder(radius, nmed, nmed, lD, size, res)
u0_single = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
                      f_sino_imag="u0_single_imag.txt",
                      f_sino_real="u0_single_real.txt")


print("Example: Backpropagation from 2D FDTD simulations")
print("Refractive index of medium:", cfg["nmed"])
print("Measurement position from object center:", cfg["lD"])
print("Wavelength sampling:", cfg["res"])
print("Performing backpropagation.")

# Set measurement parameters
# Compute scattered field from cylinder
radius = cfg["radius"]  # wavelengths
nmed = cfg["nmed"]
ncyl = cfg["ncyl"]

lD = cfg["lD"]  # measurement distance in wavelengths
lC = cfg["lC"]  # displacement from center of image
size = cfg["size"]
res = cfg["res"]  # px/wavelengths
A = cfg["A"]  # number of projections

x = np.arange(size) - size / 2.0
X, Y = np.meshgrid(x, x)
rad_px = radius * res
phantom = np.array(((Y - lC * res)**2 + X**2) < rad_px **
                   2, dtype=float) * (ncyl - nmed) + nmed

u_sinR = odt.sinogram_as_rytov(sino / u0)

# Rytov 200 projections
# remove 50 projections from total of 250 projections
remove200 = np.argsort(angles % .0002)[:50]
angles200 = np.delete(angles, remove200, axis=0)
u_sinR200 = np.delete(u_sinR, remove200, axis=0)
ph200 = unwrap_phase(np.angle(sino / u0))
ph200[remove200] = 0

fR200 = odt.backpropagate_2d(u_sinR200, angles200, res, nmed, lD*res)
nR200 = odt.odt_to_ri(fR200, res, nmed)
fR200nw = odt.backpropagate_2d(u_sinR200, angles200, res, nmed, lD*res,
                               weight_angles=False)
nR200nw = odt.odt_to_ri(fR200nw, res, nmed)

# Rytov 50 projections
remove50 = np.argsort(angles % .0002)[:200]
angles50 = np.delete(angles, remove50, axis=0)
u_sinR50 = np.delete(u_sinR, remove50, axis=0)
ph50 = unwrap_phase(np.angle(sino / u0))
ph50[remove50] = 0

fR50 = odt.backpropagate_2d(u_sinR50, angles50, res, nmed, lD*res)
nR50 = odt.odt_to_ri(fR50, res, nmed)
fR50nw = odt.backpropagate_2d(u_sinR50, angles50, res, nmed, lD*res,
                              weight_angles=False)
nR50nw = odt.odt_to_ri(fR50nw, res, nmed)

# prepare plot
kw_ri = {"vmin": 1.330,
         "vmax": 1.340}

kw_ph = {"vmin": np.min(np.array([ph200, ph50])),
         "vmax": np.max(np.array([ph200, ph50])),
         "cmap": "coolwarm",
         "interpolation": "none"}

fig, axes = plt.subplots(2, 3, figsize=(8, 4))
axes = np.array(axes).flatten()

phmap = axes[0].imshow(ph200, **kw_ph)
axes[0].set_title("Phase sinogram (200 proj.)")

rimap = axes[1].imshow(nR200nw.real, **kw_ri)
axes[1].set_title("RI without angular weights")

axes[2].imshow(nR200.real, **kw_ri)
axes[2].set_title("RI with angular weights")

axes[3].imshow(ph50, **kw_ph)
axes[3].set_title("Phase sinogram (50 proj.)")

axes[4].imshow(nR50nw.real, **kw_ri)
axes[4].set_title("RI without angular weights")

axes[5].imshow(nR50.real, **kw_ri)
axes[5].set_title("RI with angular weights")

# color bars
cbkwargs = {"fraction": 0.045,
            "format": "%.3f"}
plt.colorbar(phmap, ax=axes[0], **cbkwargs)
plt.colorbar(phmap, ax=axes[3], **cbkwargs)
plt.colorbar(rimap, ax=axes[1], **cbkwargs)
plt.colorbar(rimap, ax=axes[2], **cbkwargs)
plt.colorbar(rimap, ax=axes[5], **cbkwargs)
plt.colorbar(rimap, ax=axes[4], **cbkwargs)

plt.tight_layout()
plt.show()



"""FDTD cell phantom with tilted axis of rotation

The *in silico* data set was created with the
:abbr:`FDTD (Finite Difference Time Domain)` software `meep`_. The data
are 2D projections of a 3D refractive index phantom that is rotated
about an axis which is tilted by 0.2 rad (11.5 degrees) with respect
to the imaging plane. The example showcases the method
:func:`odtbrain.backpropagate_3d_tilted` which takes into account
such a tilted axis of rotation. The data are downsampled by a factor
of two. A total of 220 projections are used for the reconstruction.
Note that the information required for reconstruction decreases as the
tilt angle increases. If the tilt angle is 90 degrees w.r.t. the
imaging plane, then we get a rotating image of a cell (not images of a
rotating cell) and tomographic reconstruction is impossible. A brief
description of this algorithm is given in :cite:`Mueller2015tilted`.


The first column shows the measured phase, visualizing the
tilt (compare to other examples). The second column shows a
reconstruction that does not take into account the tilted axis of
rotation; the result is a blurry reconstruction. The third column
shows the improved reconstruction; the known tilted axis of rotation
is used in the reconstruction process.

.. _`meep`: http://ab-initio.mit.edu/wiki/index.php/Meep
"""
import matplotlib.pylab as plt
import numpy as np

import odtbrain as odt

from example_helper import load_data


if __name__ == "__main__":
    sino, angles, phantom, cfg = \
        load_data("fdtd_3d_sino_A220_R6.500_tiltyz0.2.tar.lzma")

    A = angles.shape[0]

    print("Example: Backpropagation from 3D FDTD simulations")
    print("Refractive index of medium:", cfg["nm"])
    print("Measurement position from object center:", cfg["lD"])
    print("Wavelength sampling:", cfg["res"])
    print("Axis tilt in y-z direction:", cfg["tilt_yz"])
    print("Number of projections:", A)

    print("Performing normal backpropagation.")
    # Apply the Rytov approximation
    sinoRytov = odt.sinogram_as_rytov(sino)

    # Perform naive backpropagation
    f_naiv = odt.backpropagate_3d(uSin=sinoRytov,
                                  angles=angles,
                                  res=cfg["res"],
                                  nm=cfg["nm"],
                                  lD=cfg["lD"]
                                  )

    print("Performing tilted backpropagation.")
    # Determine tilted axis
    tilted_axis = [0, np.cos(cfg["tilt_yz"]), np.sin(cfg["tilt_yz"])]

    # Perform tilted backpropagation
    f_tilt = odt.backpropagate_3d_tilted(uSin=sinoRytov,
                                         angles=angles,
                                         res=cfg["res"],
                                         nm=cfg["nm"],
                                         lD=cfg["lD"],
                                         tilted_axis=tilted_axis,
                                         )

    # compute refractive index n from object function
    n_naiv = odt.odt_to_ri(f_naiv, res=cfg["res"], nm=cfg["nm"])
    n_tilt = odt.odt_to_ri(f_tilt, res=cfg["res"], nm=cfg["nm"])

    sx, sy, sz = n_tilt.shape
    px, py, pz = phantom.shape

    sino_phase = np.angle(sino)

    # compare phantom and reconstruction in plot
    fig, axes = plt.subplots(2, 3, figsize=(8, 4.5))
    kwri = {"vmin": n_tilt.real.min(), "vmax": n_tilt.real.max()}
    kwph = {"vmin": sino_phase.min(), "vmax": sino_phase.max(),
            "cmap": "coolwarm"}

    # Sinogram
    axes[0, 0].set_title("phase projection")
    phmap = axes[0, 0].imshow(sino_phase[A // 2, :, :], **kwph)
    axes[0, 0].set_xlabel("detector x")
    axes[0, 0].set_ylabel("detector y")

    axes[1, 0].set_title("sinogram slice")
    axes[1, 0].imshow(sino_phase[:, :, sino.shape[2] // 2],
                      aspect=sino.shape[1] / sino.shape[0], **kwph)
    axes[1, 0].set_xlabel("detector y")
    axes[1, 0].set_ylabel("angle [rad]")
    # set y ticks for sinogram
    labels = np.linspace(0, 2 * np.pi, len(axes[1, 1].get_yticks()))
    labels = ["{:.2f}".format(i) for i in labels]
    axes[1, 0].set_yticks(np.linspace(0, len(angles), len(labels)))
    axes[1, 0].set_yticklabels(labels)

    axes[0, 1].set_title("normal (center)")
    rimap = axes[0, 1].imshow(n_naiv[sx // 2].real, **kwri)
    axes[0, 1].set_xlabel("x")
    axes[0, 1].set_ylabel("y")

    axes[1, 1].set_title("normal (nucleolus)")
    axes[1, 1].imshow(n_naiv[int(sx / 2 + 2 * cfg["res"])].real, **kwri)
    axes[1, 1].set_xlabel("x")
    axes[1, 1].set_ylabel("y")

    axes[0, 2].set_title("tilt correction (center)")
    axes[0, 2].imshow(n_tilt[sx // 2].real, **kwri)
    axes[0, 2].set_xlabel("x")
    axes[0, 2].set_ylabel("y")

    axes[1, 2].set_title("tilt correction (nucleolus)")
    axes[1, 2].imshow(n_tilt[int(sx / 2 + 2 * cfg["res"])].real, **kwri)
    axes[1, 2].set_xlabel("x")
    axes[1, 2].set_ylabel("y")

    # color bars
    cbkwargs = {"fraction": 0.045,
                "format": "%.3f"}
    plt.colorbar(phmap, ax=axes[0, 0], **cbkwargs)
    plt.colorbar(phmap, ax=axes[1, 0], **cbkwargs)
    plt.colorbar(rimap, ax=axes[0, 1], **cbkwargs)
    plt.colorbar(rimap, ax=axes[1, 1], **cbkwargs)
    plt.colorbar(rimap, ax=axes[0, 2], **cbkwargs)
    plt.colorbar(rimap, ax=axes[1, 2], **cbkwargs)

    plt.tight_layout()
    plt.show()



"""Mie cylinder with incomplete angular coverage

This example illustrates how the backpropagation algorithm of ODTbrain
handles incomplete angular coverage. All examples use 100 projections
at 100%, 60%, and 40% total angular coverage. The keyword argument
`weight_angles` that invokes angular weighting is set to `True` by
default. The *in silico* data set was created with the
softare `miefield  <https://github.com/RI-imaging/miefield>`_.
The data are 1D projections of a non-centered cylinder of constant
refractive index 1.339 embedded in water with refractive index 1.333.
The first column shows the used sinograms (missing angles are displayed
as zeros) that were created from the original sinogram with 250
projections. The second column shows the reconstruction without angular
weights and the third column shows the reconstruction with angular
weights. The keyword argument `weight_angles` was introduced in version
0.1.1.

A 180 degree coverage results in a good reconstruction of the object.
Angular weighting as implemented in the backpropagation algorithm
of ODTbrain automatically addresses uneven and incomplete angular
coverage.
"""
import matplotlib.pylab as plt
import numpy as np

import odtbrain as odt

from example_helper import load_data

sino, angles, cfg = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
                              f_angles="mie_angles.txt",
                              f_sino_real="sino_real.txt",
                              f_sino_imag="sino_imag.txt",
                              f_info="mie_info.txt")
A, size = sino.shape

# background sinogram computed with Mie theory
# miefield.GetSinogramCylinderRotation(radius, nmed, nmed, lD, lC, size, A,res)
u0 = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
               f_sino_imag="u0_imag.txt",
               f_sino_real="u0_real.txt")
# create 2d array
u0 = np.tile(u0, size).reshape(A, size).transpose()

# background field necessary to compute initial born field
# u0_single = mie.GetFieldCylinder(radius, nmed, nmed, lD, size, res)
u0_single = load_data("mie_2d_noncentered_cylinder_A250_R2.zip",
                      f_sino_imag="u0_single_imag.txt",
                      f_sino_real="u0_single_real.txt")

print("Example: Backpropagation from 2D FDTD simulations")
print("Refractive index of medium:", cfg["nmed"])
print("Measurement position from object center:", cfg["lD"])
print("Wavelength sampling:", cfg["res"])
print("Performing backpropagation.")

# Set measurement parameters
# Compute scattered field from cylinder
radius = cfg["radius"]  # wavelengths
nmed = cfg["nmed"]
ncyl = cfg["ncyl"]

lD = cfg["lD"]  # measurement distance in wavelengths
lC = cfg["lC"]  # displacement from center of image
size = cfg["size"]
res = cfg["res"]  # px/wavelengths
A = cfg["A"]  # number of projections

x = np.arange(size) - size / 2.0
X, Y = np.meshgrid(x, x)
rad_px = radius * res
phantom = np.array(((Y - lC * res)**2 + X**2) < rad_px **
                   2, dtype=float) * (ncyl - nmed) + nmed

u_sinR = odt.sinogram_as_rytov(sino / u0)

# Rytov 100 projections evenly distributed
removeeven = np.argsort(angles % .002)[:150]
angleseven = np.delete(angles, removeeven, axis=0)
u_sinReven = np.delete(u_sinR, removeeven, axis=0)
pheven = odt.sinogram_as_radon(sino / u0)
pheven[removeeven] = 0

fReven = odt.backpropagate_2d(u_sinReven, angleseven, res, nmed, lD * res)
nReven = odt.odt_to_ri(fReven, res, nmed)
fRevennw = odt.backpropagate_2d(
    u_sinReven, angleseven, res, nmed, lD * res, weight_angles=False)
nRevennw = odt.odt_to_ri(fRevennw, res, nmed)

# Rytov 100 projections more than 180
removemiss = 249 - \
    np.concatenate((np.arange(100), 100 + np.arange(150)[::3]))
anglesmiss = np.delete(angles, removemiss, axis=0)
u_sinRmiss = np.delete(u_sinR, removemiss, axis=0)
phmiss = odt.sinogram_as_radon(sino / u0)
phmiss[removemiss] = 0

fRmiss = odt.backpropagate_2d(u_sinRmiss, anglesmiss, res, nmed, lD * res)
nRmiss = odt.odt_to_ri(fRmiss, res, nmed)
fRmissnw = odt.backpropagate_2d(
    u_sinRmiss, anglesmiss, res, nmed, lD * res, weight_angles=False)
nRmissnw = odt.odt_to_ri(fRmissnw, res, nmed)

# Rytov 100 projections less than 180
removebad = 249 - np.arange(150)
anglesbad = np.delete(angles, removebad, axis=0)
u_sinRbad = np.delete(u_sinR, removebad, axis=0)
phbad = odt.sinogram_as_radon(sino / u0)
phbad[removebad] = 0

fRbad = odt.backpropagate_2d(u_sinRbad, anglesbad, res, nmed, lD * res)
nRbad = odt.odt_to_ri(fRbad, res, nmed)
fRbadnw = odt.backpropagate_2d(
    u_sinRbad, anglesbad, res, nmed, lD * res, weight_angles=False)
nRbadnw = odt.odt_to_ri(fRbadnw, res, nmed)

# prepare plot
kw_ri = {"vmin": np.min(np.array([phantom, nRmiss.real, nReven.real])),
         "vmax": np.max(np.array([phantom, nRmiss.real, nReven.real]))}

kw_ph = {"vmin": np.min(np.array([pheven, phmiss])),
         "vmax": np.max(np.array([pheven, phmiss])),
         "cmap": "coolwarm",
         "interpolation": "none"}

fig, axes = plt.subplots(3, 3, figsize=(8, 6.5))

axes[0, 0].set_title("100% coverage ({} proj.)".format(angleseven.shape[0]))
phmap = axes[0, 0].imshow(pheven, **kw_ph)

axes[0, 1].set_title("RI without angular weights")
rimap = axes[0, 1].imshow(nRevennw.real, **kw_ri)

axes[0, 2].set_title("RI with angular weights")
rimap = axes[0, 2].imshow(nReven.real, **kw_ri)

axes[1, 0].set_title("60% coverage ({} proj.)".format(anglesmiss.shape[0]))
axes[1, 0].imshow(phmiss, **kw_ph)

axes[1, 1].set_title("RI without angular weights")
axes[1, 1].imshow(nRmissnw.real, **kw_ri)

axes[1, 2].set_title("RI with angular weights")
axes[1, 2].imshow(nRmiss.real, **kw_ri)

axes[2, 0].set_title("40% coverage ({} proj.)".format(anglesbad.shape[0]))
axes[2, 0].imshow(phbad, **kw_ph)

axes[2, 1].set_title("RI without angular weights")
axes[2, 1].imshow(nRbadnw.real, **kw_ri)

axes[2, 2].set_title("RI with angular weights")
axes[2, 2].imshow(nRbad.real, **kw_ri)

# color bars
cbkwargs = {"fraction": 0.045,
            "format": "%.3f"}
plt.colorbar(phmap, ax=axes[0, 0], **cbkwargs)
plt.colorbar(phmap, ax=axes[1, 0], **cbkwargs)
plt.colorbar(phmap, ax=axes[2, 0], **cbkwargs)
plt.colorbar(rimap, ax=axes[0, 1], **cbkwargs)
plt.colorbar(rimap, ax=axes[1, 1], **cbkwargs)
plt.colorbar(rimap, ax=axes[2, 1], **cbkwargs)
plt.colorbar(rimap, ax=axes[0, 2], **cbkwargs)
plt.colorbar(rimap, ax=axes[1, 2], **cbkwargs)
plt.colorbar(rimap, ax=axes[2, 2], **cbkwargs)

plt.tight_layout()
plt.show()



"""Sinogram preparation"""
import numpy as np
from scipy.stats import mode
from skimage.restoration import unwrap_phase


def align_unwrapped(sino):
    """Align an unwrapped phase array to zero-phase

    All operations are performed in-place.
    """
    samples = []
    if len(sino.shape) == 2:
        # 2D
        # take 1D samples at beginning and end of array
        samples.append(sino[:, 0])
        samples.append(sino[:, 1])
        samples.append(sino[:, 2])
        samples.append(sino[:, -1])
        samples.append(sino[:, -2])

    elif len(sino.shape) == 3:
        # 3D
        # take 1D samples at beginning and end of array
        samples.append(sino[:, 0, 0])
        samples.append(sino[:, 0, -1])
        samples.append(sino[:, -1, 0])
        samples.append(sino[:, -1, -1])
        samples.append(sino[:, 0, 1])

    # find discontinuities in the samples
    steps = np.zeros((len(samples), samples[0].shape[0]))
    for i in range(len(samples)):
        t = np.unwrap(samples[i])
        steps[i] = samples[i] - t

    # if the majority believes so, add a step of PI
    remove = mode(steps, axis=0, keepdims=True)[0][0]

    # obtain divmod min
    twopi = 2*np.pi
    minimum = divmod_neg(np.min(sino), twopi)[0]
    remove += minimum*twopi

    for i in range(len(sino)):
        sino[i] -= remove[i]


def divmod_neg(a, b):
    """Return divmod with closest result to zero"""
    q, r = divmod(a, b)
    # make sure r is close to zero
    sr = np.sign(r)
    if np.abs(r) > b/2:
        q += sr
        r -= b * sr
    return q, r


def sinogram_as_radon(uSin, align=True):
    r"""Compute the phase from a complex wave field sinogram

    This step is essential when using the ray approximation before
    computation of the refractive index with the inverse Radon
    transform.

    Parameters
    ----------
    uSin: 2d or 3d complex ndarray
        The background-corrected sinogram of the complex scattered wave
        :math:`u(\mathbf{r})/u_0(\mathbf{r})`. The first axis iterates
        through the angles :math:`\phi_0`.
    align: bool
        Tries to correct for a phase offset in the phase sinogram.

    Returns
    -------
    phase: 2d or 3d real ndarray
        The unwrapped phase array corresponding to `uSin`.

    See Also
    --------
    skimage.restoration.unwrap_phase: phase unwrapping
    radontea.backproject_3d: e.g. reconstruction via backprojection
    """
    ndims = len(uSin.shape)

    if ndims == 2:
        # unwrapping is very important
        phiR = np.unwrap(np.angle(uSin), axis=-1)
    else:
        # Unwrap gets the dimension of the problem from the input
        # data. Since we have a sinogram, we need to pass it the
        # slices one by one.
        phiR = np.angle(uSin)
        for ii in range(len(phiR)):
            phiR[ii] = unwrap_phase(phiR[ii], rng=47)

    if align:
        align_unwrapped(phiR)

    return phiR


def sinogram_as_rytov(uSin, u0=1, align=True):
    r"""Convert the complex wave field sinogram to the Rytov phase

    This method applies the Rytov approximation to the
    recorded complex wave sinogram. To achieve this, the following
    filter is applied:

    .. math::
        u_\mathrm{B}(\mathbf{r}) = u_\mathrm{0}(\mathbf{r})
            \ln\!\left(
            \frac{u_\mathrm{R}(\mathbf{r})}{u_\mathrm{0}(\mathbf{r})}
             +1 \right)

    This filter step effectively replaces the Born approximation
    :math:`u_\mathrm{B}(\mathbf{r})` with the Rytov approximation
    :math:`u_\mathrm{R}(\mathbf{r})`, assuming that the scattered
    field is equal to
    :math:`u(\mathbf{r})\approx u_\mathrm{R}(\mathbf{r})+
    u_\mathrm{0}(\mathbf{r})`.


    Parameters
    ----------
    uSin: 2d or 3d complex ndarray
        The sinogram of the complex wave
        :math:`u_\mathrm{R}(\mathbf{r}) + u_\mathrm{0}(\mathbf{r})`.
        The first axis iterates through the angles :math:`\phi_0`.
    u0: ndarray of dimension as `uSin` or less, or int.
        The incident plane wave
        :math:`u_\mathrm{0}(\mathbf{r})` at the detector.
        If `u0` is "1", it is assumed that the data is already
        background-corrected (
        `uSin` :math:`= \frac{u_\mathrm{R}(\mathbf{r})}{
        u_\mathrm{0}(\mathbf{r})} + 1`
        ). Note that if the reconstruction distance :math:`l_\mathrm{D}`
        of the original experiment is non-zero and `u0` is set to 1,
        then the reconstruction will be wrong; the field is not focused
        to the center of the reconstruction volume.
    align: bool
        Tries to correct for a phase offset in the phase sinogram.

    Returns
    -------
    uB: 2d or 3d real ndarray
        The Rytov-filtered complex sinogram
        :math:`u_\mathrm{B}(\mathbf{r})`.

    See Also
    --------
    skimage.restoration.unwrap_phase: phase unwrapping
    """
    ndims = len(uSin.shape)

    # imaginary part of the complex Rytov phase
    phiR = np.angle(uSin / u0)

    # real part of the complex Rytov phase
    lna = np.log(np.absolute(uSin / u0))

    if ndims == 2:
        # unwrapping is very important
        phiR[:] = np.unwrap(phiR, axis=-1)
    else:
        # Unwrap gets the dimension of the problem from the input
        # data. Since we have a sinogram, we need to pass it the
        # slices one by one.
        for ii in range(len(phiR)):
            phiR[ii] = unwrap_phase(phiR[ii], rng=47)

    if align:
        align_unwrapped(phiR)

    # rytovSin = u0*(np.log(a/a0) + 1j*phiR)
    # u0 is one - we already did background correction

    # complex rytov phase:
    rytovSin = 1j * phiR + lna
    return u0 * rytovSin



"""Missing apple core correction

The missing apple core :cite:`Vertu2009` is a phenomenon in diffraction
tomography that is a result of the fact the the Fourier space is not
filled completely when the sample is rotated only about a single axis.
The resulting artifacts include ringing and blurring in the
reconstruction parallel to the original rotation axis. By enforcing
constraints (refractive index real-valued and larger than the
surrounding medium), these artifacts can be attenuated.

This example generates an artificial sinogram using the Python
library :ref:`cellsino <cellsino:index>` (The example parameters
are reused from :ref:`this example <cellsino:example_simple_cell>`).
The sinogram is then reconstructed with the backpropagation algorithm
and the missing apple core correction is applied.

.. note::
    The missing apple core correction :func:`odtbrain.apple.correct`
    was implemented in version 0.3.0 and is thus not used in the
    older examples.
"""
import matplotlib.pylab as plt
import numpy as np

import cellsino
import odtbrain as odt


if __name__ == "__main__":
    # number of sinogram angles
    num_ang = 160
    # sinogram acquisition angles
    angles = np.linspace(0, 2*np.pi, num_ang, endpoint=False)
    # detector grid size
    grid_size = (250, 250)
    # vacuum wavelength [m]
    wavelength = 550e-9
    # pixel size [m]
    pixel_size = 0.08e-6
    # refractive index of the surrounding medium
    medium_index = 1.335

    # initialize cell phantom
    phantom = cellsino.phantoms.SimpleCell()

    # initialize sinogram with geometric parameters
    sino = cellsino.Sinogram(phantom=phantom,
                             wavelength=wavelength,
                             pixel_size=pixel_size,
                             grid_size=grid_size)

    # compute sinogram (field with Rytov approximation and fluorescence)
    sino = sino.compute(angles=angles, propagator="rytov", mode="field")

    # reconstruction of refractive index
    sino_rytov = odt.sinogram_as_rytov(sino)
    f = odt.backpropagate_3d(uSin=sino_rytov,
                             angles=angles,
                             res=wavelength/pixel_size,
                             nm=medium_index)

    ri = odt.odt_to_ri(f=f,
                       res=wavelength/pixel_size,
                       nm=medium_index)

    # apple core correction
    fc = odt.apple.correct(f=f,
                           res=wavelength/pixel_size,
                           nm=medium_index,
                           method="sh")

    ric = odt.odt_to_ri(f=fc,
                        res=wavelength/pixel_size,
                        nm=medium_index)

    # plotting
    idx = ri.shape[2] // 2

    # log-scaled power spectra
    ft = np.log(1 + np.abs(np.fft.fftshift(np.fft.fftn(ri))))
    ftc = np.log(1 + np.abs(np.fft.fftshift(np.fft.fftn(ric))))

    plt.figure(figsize=(7, 5.5))

    plotkwri = {"vmax": ri.real.max(),
                "vmin": ri.real.min(),
                "interpolation": "none",
                }

    plotkwft = {"vmax": ft.max(),
                "vmin": 0,
                "interpolation": "none",
                }

    ax1 = plt.subplot(221, title="plain refractive index")
    mapper = ax1.imshow(ri[:, :, idx].real, **plotkwri)
    plt.colorbar(mappable=mapper, ax=ax1)

    ax2 = plt.subplot(222, title="corrected refractive index")
    mapper = ax2.imshow(ric[:, :, idx].real, **plotkwri)
    plt.colorbar(mappable=mapper, ax=ax2)

    ax3 = plt.subplot(223, title="Fourier space (visible apple core)")
    mapper = ax3.imshow(ft[:, :, idx], **plotkwft)
    plt.colorbar(mappable=mapper, ax=ax3)

    ax4 = plt.subplot(224, title="Fourier space (with correction)")
    mapper = ax4.imshow(ftc[:, :, idx], **plotkwft)
    plt.colorbar(mappable=mapper, ax=ax4)

    plt.tight_layout()
    plt.show()



"""FDTD cell phantom with tilted and rolled axis of rotation

The *in silico* data set was created with the
:abbr:`FDTD (Finite Difference Time Domain)` software `meep`_. The data
are 2D projections of a 3D refractive index phantom that is rotated
about an axis which is tilted by 0.2 rad (11.5 degrees) with respect to
the imaging plane and rolled by -.42 rad (-24.1 degrees) within the
imaging plane. The data are the same as were used in the previous
example. A brief description of this algorithm is given in
:cite:`Mueller2015tilted`.

.. _`meep`: http://ab-initio.mit.edu/wiki/index.php/Meep
"""
import matplotlib.pylab as plt
import numpy as np
from scipy.ndimage import rotate

import odtbrain as odt

from example_helper import load_data


if __name__ == "__main__":
    sino, angles, phantom, cfg = \
        load_data("fdtd_3d_sino_A220_R6.500_tiltyz0.2.tar.lzma")

    # Perform titlt by -.42 rad in detector plane
    rotang = -0.42
    rotkwargs = {"mode": "constant",
                 "order": 2,
                 "reshape": False,
                 }
    for ii in range(len(sino)):
        sino[ii].real = rotate(
            sino[ii].real, np.rad2deg(rotang), cval=1, **rotkwargs)
        sino[ii].imag = rotate(
            sino[ii].imag, np.rad2deg(rotang), cval=0, **rotkwargs)

    A = angles.shape[0]

    print("Example: Backpropagation from 3D FDTD simulations")
    print("Refractive index of medium:", cfg["nm"])
    print("Measurement position from object center:", cfg["lD"])
    print("Wavelength sampling:", cfg["res"])
    print("Axis tilt in y-z direction:", cfg["tilt_yz"])
    print("Number of projections:", A)

    # Apply the Rytov approximation
    sinoRytov = odt.sinogram_as_rytov(sino)

    # Determine tilted axis
    tilted_axis = [0, np.cos(cfg["tilt_yz"]), np.sin(cfg["tilt_yz"])]
    rotmat = np.array([
        [np.cos(rotang), -np.sin(rotang), 0],
        [np.sin(rotang), np.cos(rotang), 0],
        [0, 0, 1],
    ])
    tilted_axis = np.dot(rotmat, tilted_axis)

    print("Performing tilted backpropagation.")
    # Perform tilted backpropagation
    f_tilt = odt.backpropagate_3d_tilted(uSin=sinoRytov,
                                         angles=angles,
                                         res=cfg["res"],
                                         nm=cfg["nm"],
                                         lD=cfg["lD"],
                                         tilted_axis=tilted_axis,
                                         )

    # compute refractive index n from object function
    n_tilt = odt.odt_to_ri(f_tilt, res=cfg["res"], nm=cfg["nm"])

    sx, sy, sz = n_tilt.shape
    px, py, pz = phantom.shape

    sino_phase = np.angle(sino)

    # compare phantom and reconstruction in plot
    fig, axes = plt.subplots(1, 3, figsize=(8, 2.4))
    kwri = {"vmin": n_tilt.real.min(), "vmax": n_tilt.real.max()}
    kwph = {"vmin": sino_phase.min(), "vmax": sino_phase.max(),
            "cmap": "coolwarm"}

    # Sinogram
    axes[0].set_title("phase projection")
    phmap = axes[0].imshow(sino_phase[A // 2, :, :], **kwph)
    axes[0].set_xlabel("detector x")
    axes[0].set_ylabel("detector y")

    axes[1].set_title("sinogram slice")
    axes[1].imshow(sino_phase[:, :, sino.shape[2] // 2],
                   aspect=sino.shape[1] / sino.shape[0], **kwph)
    axes[1].set_xlabel("detector y")
    axes[1].set_ylabel("angle [rad]")
    # set y ticks for sinogram
    labels = np.linspace(0, 2 * np.pi, len(axes[1].get_yticks()))
    labels = ["{:.2f}".format(i) for i in labels]
    axes[1].set_yticks(np.linspace(0, len(angles), len(labels)))
    axes[1].set_yticklabels(labels)

    axes[2].set_title("tilt correction (nucleolus)")
    rimap = axes[2].imshow(n_tilt[int(sx / 2 + 2 * cfg["res"])].real, **kwri)
    axes[2].set_xlabel("x")
    axes[2].set_ylabel("y")

    # color bars
    cbkwargs = {"fraction": 0.045,
                "format": "%.3f"}
    plt.colorbar(phmap, ax=axes[0], **cbkwargs)
    plt.colorbar(phmap, ax=axes[1], **cbkwargs)
    plt.colorbar(rimap, ax=axes[2], **cbkwargs)

    plt.tight_layout()
    plt.show()



"""Test sinogram weighting"""
import numpy as np
import platform

import odtbrain

from common_methods import create_test_sino_3d, get_test_parameter_set


def test_3d_backprop_weights_even():
    """
    even weights
    """
    platform.system = lambda: "Windows"
    sino, angles = create_test_sino_3d()
    p = get_test_parameter_set(1)[0]
    f1 = odtbrain.backpropagate_3d(sino, angles, weight_angles=False, **p)
    f2 = odtbrain.backpropagate_3d(sino, angles, weight_angles=True, **p)
    data1 = np.array(f1).flatten().view(float)
    data2 = np.array(f2).flatten().view(float)
    assert np.allclose(data1, data2)


def test_3d_backprop_tilted_weights_even():
    """
    even weights
    """
    platform.system = lambda: "Windows"
    sino, angles = create_test_sino_3d()
    p = get_test_parameter_set(1)[0]
    f1 = odtbrain.backpropagate_3d_tilted(
        sino, angles, weight_angles=False, **p)
    f2 = odtbrain.backpropagate_3d_tilted(
        sino, angles, weight_angles=True, **p)
    data1 = np.array(f1).flatten().view(float)
    data2 = np.array(f2).flatten().view(float)
    assert np.allclose(data1, data2)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test copying arrays"""
import numpy as np

import odtbrain

from common_methods import create_test_sino_3d, get_test_parameter_set


def test_back3d():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    # complex
    f1 = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                   dtype=float,
                                   copy=False, **p)
    f2 = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                   dtype=float,
                                   copy=True, **p)
    assert np.allclose(f1, f2)


def test_back3d_tilted():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    f1 = odtbrain.backpropagate_3d_tilted(sino, angles, padval=0,
                                          dtype=float,
                                          copy=False, **p)
    f2 = odtbrain.backpropagate_3d_tilted(sino, angles, padval=0,
                                          dtype=float,
                                          copy=True, **p)
    assert np.allclose(f1, f2)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""2D backpropagation algorithm"""
import numpy as np
import scipy.ndimage

from . import util


def backpropagate_2d(uSin, angles, res, nm, lD=0, coords=None,
                     weight_angles=True,
                     onlyreal=False, padding=True, padval=0,
                     count=None, max_count=None, verbose=0):
    r"""2D backpropagation with the Fourier diffraction theorem

    Two-dimensional diffraction tomography reconstruction
    algorithm for scattering of a plane wave
    :math:`u_0(\mathbf{r}) = u_0(x,z)`
    by a dielectric object with refractive index
    :math:`n(x,z)`.

    This method implements the 2D backpropagation algorithm
    :cite:`Mueller2015arxiv`.

    .. math::
        f(\mathbf{r}) =
            -\frac{i k_\mathrm{m}}{2\pi}
            \sum_{j=1}^{N} \! \Delta \phi_0 D_{-\phi_j} \!\!
            \left \{
            \text{FFT}^{-1}_{\mathrm{1D}}
            \left \{
            \left| k_\mathrm{Dx} \right|
            \frac{\text{FFT}_{\mathrm{1D}} \left \{
            u_{\mathrm{B},\phi_j}(x_\mathrm{D}) \right \}
            }{u_0(l_\mathrm{D})}
            \exp \! \left[i k_\mathrm{m}(M - 1) \cdot
            (z_{\phi_j}-l_\mathrm{D}) \right]
            \right \}
            \right \}

    with the forward :math:`\text{FFT}_{\mathrm{1D}}` and inverse
    :math:`\text{FFT}^{-1}_{\mathrm{1D}}` 1D fast Fourier transform, the
    rotational operator :math:`D_{-\phi_j}`, the angular distance between the
    projections :math:`\Delta \phi_0`, the ramp filter in Fourier space
    :math:`|k_\mathrm{Dx}|`, and the propagation distance
    :math:`(z_{\phi_j}-l_\mathrm{D})`.

    Parameters
    ----------
    uSin: (A,N) ndarray
        Two-dimensional sinogram of line recordings
        :math:`u_{\mathrm{B}, \phi_j}(x_\mathrm{D})`
        divided by the incident plane wave :math:`u_0(l_\mathrm{D})`
        measured at the detector.
    angles: (A,) ndarray
        Angular positions :math:`\phi_j` of `uSin` in radians.
    res: float
        Vacuum wavelength of the light :math:`\lambda` in pixels.
    nm: float
        Refractive index of the surrounding medium :math:`n_\mathrm{m}`.
    lD: float
        Distance from center of rotation to detector plane
        :math:`l_\mathrm{D}` in pixels.
    coords: None [(2,M) ndarray]
        Computes only the output image at these coordinates. This
        keyword is reserved for future versions and is not
        implemented yet.
    weight_angles: bool
        If `True`, weights each backpropagated projection with a factor
        proportional to the angular distance between the neighboring
        projections.

        .. math::
            \Delta \phi_0 \longmapsto \Delta \phi_j =
                \frac{\phi_{j+1} - \phi_{j-1}}{2}

        .. versionadded:: 0.1.1
    onlyreal: bool
        If `True`, only the real part of the reconstructed image
        will be returned. This saves computation time.
    padding: bool
        Pad the input data to the second next power of 2 before
        Fourier transforming. This reduces artifacts and speeds up
        the process for input image sizes that are not powers of 2.
    padval: float
        The value used for padding. This is important for the Rytov
        approximation, where an approximate zero in the phase might
        translate to 2πi due to the unwrapping algorithm. In that
        case, this value should be a multiple of 2πi.
        If `padval` is `None`, then the edge values are used for
        padding (see documentation of :func:`numpy.pad`).
    count, max_count: multiprocessing.Value or `None`
        Can be used to monitor the progress of the algorithm.
        Initially, the value of `max_count.value` is incremented
        by the total number of steps. At each step, the value
        of `count.value` is incremented.
    verbose: int
        Increment to increase verbosity.


    Returns
    -------
    f: ndarray of shape (N,N), complex if `onlyreal` is `False`
        Reconstructed object function :math:`f(\mathbf{r})` as defined
        by the Helmholtz equation.
        :math:`f(x,z) =
        k_m^2 \left(\left(\frac{n(x,z)}{n_m}\right)^2 -1\right)`


    See Also
    --------
    odt_to_ri: conversion of the object function :math:`f(\mathbf{r})`
        to refractive index :math:`n(\mathbf{r})`

    radontea.backproject: backprojection based on the Fourier slice
        theorem

    Notes
    -----
    Do not use the parameter `lD` in combination with the Rytov
    approximation - the propagation is not correctly described.
    Instead, numerically refocus the sinogram prior to converting
    it to Rytov data (using e.g. :func:`odtbrain.sinogram_as_rytov`)
    with a numerical focusing algorithm (available in the Python
    package :py:mod:`nrefocus`).
    """
    ##
    ##
    # TODO:
    # - combine the 2nd filter and the rotation in the for loop
    # to save memory. However, memory is not a big issue in 2D.
    ##
    ##
    A = angles.shape[0]
    if max_count is not None:
        max_count.value += A + 2
    # Check input data
    assert len(uSin.shape) == 2, "Input data `uB` must have shape (A,N)!"
    assert len(uSin) == A, "`len(angles)` must be  equal to `len(uSin)`!"

    if coords is not None:
        raise NotImplementedError("Output coordinates cannot yet be set " +
                                  + "for the 2D backrpopagation algorithm.")
    # Cut-Off frequency
    # km [1/px]
    km = (2 * np.pi * nm) / res
    # Here, the notation defines
    # a wave propagating to the right as:
    #
    #    u0(x) = exp(ikx)
    #
    # However, in physics usually we use the other sign convention:
    #
    #    u0(x) = exp(-ikx)
    #
    # In order to be consistent with programs like Meep or our
    # scattering script for a dielectric cylinder, we want to use the
    # latter sign convention.
    # This is not a big problem. We only need to multiply the imaginary
    # part of the scattered wave by -1.

    # Perform weighting
    if weight_angles:
        weights = util.compute_angle_weights_1d(angles).reshape(-1, 1)
        sinogram = uSin * weights
    else:
        sinogram = uSin

    # Size of the input data
    ln = sinogram.shape[1]

    # We perform padding before performing the Fourier transform.
    # This gets rid of artifacts due to false periodicity and also
    # speeds up Fourier transforms of the input image size is not
    # a power of 2.
    order = max(64., 2**np.ceil(np.log(ln * 2.1) / np.log(2)))

    if padding:
        pad = order - ln
    else:
        pad = 0

    padl = int(np.ceil(pad / 2))
    padr = int(pad - padl)

    if padval is None:
        sino = np.pad(sinogram, ((0, 0), (padl, padr)),
                      mode="edge")
        if verbose > 0:
            print("......Padding with edge values.")
    else:
        sino = np.pad(sinogram, ((0, 0), (padl, padr)),
                      mode="linear_ramp",
                      end_values=(padval,))
        if verbose > 0:
            print("......Verifying padding value: {}".format(padval))

    # zero-padded length of sinogram.
    lN = sino.shape[1]

    # Ask for the filter. Do not include zero (first element).
    #
    # Integrals over ϕ₀ [0,2π]; kx [-kₘ,kₘ]
    #   - double coverage factor 1/2 already included
    #   - unitary angular frequency to unitary ordinary frequency
    #     conversion performed in calculation of UB=FT(uB).
    #
    # f(r) = -i kₘ / ((2π)^(3/2) a₀)            (prefactor)
    #      * iint dϕ₀ dkx                       (prefactor)
    #      * |kx|                               (prefactor)
    #      * exp(-i kₘ M lD )                   (prefactor)
    #      * UBϕ₀(kx)                             (dependent on ϕ₀)
    #      * exp( i (kx t⊥ + kₘ (M - 1) s₀) r )   (dependent on ϕ₀ and r)
    #
    # (r and s₀ are vectors. In the last term we perform the dot-product)
    #
    # kₘM = sqrt( kₘ² - kx² )
    # t⊥  = (  cos(ϕ₀), sin(ϕ₀) )
    # s₀  = ( -sin(ϕ₀), cos(ϕ₀) )
    #
    # The filter can be split into two parts
    #
    # 1) part without dependence on the z-coordinate
    #
    #        -i kₘ / ((2π)^(3/2) a₀)
    #      * iint dϕ₀ dkx
    #      * |kx|
    #      * exp(-i kₘ M lD )
    #
    # 2) part with dependence of the z-coordinate
    #
    #        exp( i (kx t⊥ + kₘ (M - 1) s₀) r )
    #
    # The filter (1) can be performed using the classical filter process
    # as in the backprojection algorithm.
    #
    #
    if count is not None:
        count.value += 1

    # Corresponding sample frequencies
    fx = np.fft.fftfreq(lN)  # 1D array
    # kx is a 1D array.
    kx = 2 * np.pi * fx
    # Differentials for integral
    dphi0 = 2 * np.pi / A
    # We will later multiply with phi0.
    #               a, x
    kx = kx.reshape(1, -1)
    # Low-pass filter:
    # less-than-or-equal would give us zero division error.
    filter_klp = (kx**2 < km**2)

    # Filter M so there are no nans from the root
    M = 1. / km * np.sqrt((km**2 - kx**2) * filter_klp)

    prefactor = -1j * km / (2 * np.pi)
    prefactor *= dphi0
    prefactor *= np.abs(kx) * filter_klp
    # new in version 0.1.4:
    # We multiply by the factor (M-1) instead of just (M)
    # to take into account that we have a scattered
    # wave that is normalized by u0.
    prefactor *= np.exp(-1j * km * (M-1) * lD)
    # Perform filtering of the sinogram
    projection = np.fft.fft(sino, axis=-1) * prefactor

    #
    # filter (2) must be applied before rotation as well
    # exp( i (kx t⊥ + kₘ (M - 1) s₀) r )
    #
    # t⊥  = (  cos(ϕ₀), sin(ϕ₀) )
    # s₀  = ( -sin(ϕ₀), cos(ϕ₀) )
    #
    # This filter is effectively an inverse Fourier transform
    #
    # exp(i kx xD) exp(i kₘ (M - 1) yD )
    #
    # xD =   x cos(ϕ₀) + y sin(ϕ₀)
    # yD = - x sin(ϕ₀) + y cos(ϕ₀)

    # Everything is in pixels
    center = ln / 2.0
    x = np.arange(lN) - center + .5
    # Meshgrid for output array
    yv = x.reshape(-1, 1)

    Mp = M.reshape(1, -1)
    filter2 = np.exp(1j * yv * km * (Mp - 1))  # .reshape(1,lN,lN)

    projection = projection.reshape(A, 1, lN)  # * filter2

    # Prepare complex output image
    if onlyreal:
        outarr = np.zeros((ln, ln))
    else:
        outarr = np.zeros((ln, ln), dtype=np.dtype(complex))

    if count is not None:
        count.value += 1

    # Calculate backpropagations
    for i in np.arange(A):
        # Create an interpolation object of the projection.

        # interpolation of the rotated fourier transformed projection
        # this is already tiled onto the entire image.
        sino_filtered = np.fft.ifft(projection[i] * filter2, axis=-1)

        # Resize filtered sinogram back to original size
        sino = sino_filtered[:ln, padl:padl + ln]

        rotated_projr = scipy.ndimage.rotate(
            sino.real, -angles[i] * 180 / np.pi,
            reshape=False, mode="constant", cval=0)
        # Append results

        outarr += rotated_projr

        if not onlyreal:
            outarr += 1j * scipy.ndimage.rotate(
                sino.imag, -angles[i] * 180 / np.pi,
                reshape=False, mode="constant", cval=0)

        if count is not None:
            count.value += 1

    return outarr



"""Test tilted backpropagation algorithm"""
import numpy as np

import odtbrain

from common_methods import create_test_sino_3d, create_test_sino_3d_tilted, \
    cutout, get_test_parameter_set


def test_3d_backprop_phase_real():
    sino, angles = create_test_sino_3d()
    parameters = get_test_parameter_set(2)
    # reference
    rref = list()
    for p in parameters:
        fref = odtbrain.backpropagate_3d(sino, angles, padval=0,
                                         dtype=float, onlyreal=True, **p)
        rref.append(cutout(fref))
    dataref = np.array(rref).flatten().view(float)

    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d_tilted(sino, angles, padval=0,
                                             dtype=float, onlyreal=True,
                                             **p)
        r.append(cutout(f))
    data = np.array(r).flatten().view(float)
    assert np.allclose(data, dataref)


def test_3d_backprop_pad():
    sino, angles = create_test_sino_3d()
    parameters = get_test_parameter_set(2)
    # reference
    rref = list()
    for p in parameters:
        fref = odtbrain.backpropagate_3d(sino, angles, padval="edge",
                                         dtype=float, onlyreal=False, **p)
        rref.append(cutout(fref))
    dataref = np.array(rref).flatten().view(float)

    r = list()
    for p in parameters:
        f = odtbrain.backpropagate_3d_tilted(sino, angles, padval="edge",
                                             dtype=float, onlyreal=False,
                                             **p)
        r.append(cutout(f))
    data = np.array(r).flatten().view(float)

    assert np.allclose(data, dataref)


def test_3d_backprop_plane_rotation():
    """
    A very soft test to check if planar rotation works fine
    in the reconstruction with tilted angles.
    """
    parameters = get_test_parameter_set(1)
    results = []

    # These are specially selected angles that don't give high results.
    # Probably due to phase-wrapping, errors >2 may appear. Hence, we
    # call it a soft test.
    tilts = [1.1, 0.0, 0.234, 2.80922, -.29, 9.87]

    for angz in tilts:
        sino, angles = create_test_sino_3d_tilted(tilt_plane=angz, A=21)
        rotmat = np.array([
            [np.cos(angz), -np.sin(angz), 0],
            [np.sin(angz),  np.cos(angz), 0],
            [0,             0, 1],
        ])
        # rotate `tilted_axis` onto the y-z plane.
        tilted_axis = np.dot(rotmat, [0, 1, 0])

        rref = list()
        for p in parameters:
            fref = odtbrain.backpropagate_3d_tilted(sino, angles,
                                                    padval="edge",
                                                    tilted_axis=tilted_axis,
                                                    padding=(False, False),
                                                    dtype=float,
                                                    onlyreal=False,
                                                    **p)
            rref.append(cutout(fref))
        data = np.array(rref).flatten().view(float)
        results.append(data)

    for ii in np.arange(len(results)):
        assert np.allclose(results[ii], results[ii-1], atol=.2, rtol=.2)


def test_3d_backprop_plane_alignment_along_axes():
    """
    Tests whether the reconstruction is always aligned with
    the rotational axis (and not antiparallel).
    """
    parameters = get_test_parameter_set(1)
    p = parameters[0]
    results = []

    # These are specially selected angles that don't give high results.
    # Probably due to phase-wrapping, errors >2 may appear. Hence, we
    # call it a soft test.
    tilts = [0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi]

    for angz in tilts:
        sino, angles = create_test_sino_3d_tilted(tilt_plane=angz, A=21)
        rotmat = np.array([
            [np.cos(angz), -np.sin(angz), 0],
            [np.sin(angz),  np.cos(angz), 0],
            [0,             0, 1],
        ])
        # rotate `tilted_axis` onto the y-z plane.
        tilted_axis = np.dot(rotmat, [0, 1, 0])
        fref = odtbrain.backpropagate_3d_tilted(sino, angles,
                                                padval="edge",
                                                tilted_axis=tilted_axis,
                                                padding=(False, False),
                                                dtype=float,
                                                onlyreal=True,
                                                **p)
        results.append(fref)

    for ii in np.arange(len(results)):
        assert np.allclose(results[ii], results[ii-1], atol=.2, rtol=.2)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Tests progress counters"""
import multiprocessing as mp

import pytest

import odtbrain

from common_methods import create_test_sino_2d, create_test_sino_3d, \
    get_test_parameter_set


@pytest.mark.filterwarnings(
    "ignore::odtbrain.warn.DataUndersampledWarning")
def test_integrate_2d():
    sino, angles = create_test_sino_2d(N=10)
    p = get_test_parameter_set(1)[0]
    # complex
    jmc = mp.Value("i", 0)
    jmm = mp.Value("i", 0)

    odtbrain.integrate_2d(sino, angles,
                          count=jmc,
                          max_count=jmm,
                          **p)

    assert jmc.value == jmm.value
    assert jmc.value != 0


@pytest.mark.filterwarnings(
    "ignore::odtbrain.warn.ImplementationAmbiguousWarning")
def test_fmp_2d():
    sino, angles = create_test_sino_2d(N=10)
    p = get_test_parameter_set(1)[0]
    # complex
    jmc = mp.Value("i", 0)
    jmm = mp.Value("i", 0)

    odtbrain.fourier_map_2d(sino, angles,
                            count=jmc,
                            max_count=jmm,
                            **p)

    assert jmc.value == jmm.value
    assert jmc.value != 0


def test_bpp_2d():
    sino, angles = create_test_sino_2d(N=10)
    p = get_test_parameter_set(1)[0]
    # complex
    jmc = mp.Value("i", 0)
    jmm = mp.Value("i", 0)

    odtbrain.backpropagate_2d(sino, angles, padval=0,
                              count=jmc,
                              max_count=jmm,
                              **p)
    assert jmc.value == jmm.value
    assert jmc.value != 0


def test_back3d():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    # complex
    jmc = mp.Value("i", 0)
    jmm = mp.Value("i", 0)
    odtbrain.backpropagate_3d(sino, angles, padval=0,
                              dtype=float,
                              count=jmc,
                              max_count=jmm,
                              **p)
    assert jmc.value == jmm.value
    assert jmc.value != 0


def test_back3d_tilted():
    sino, angles = create_test_sino_3d(Nx=10, Ny=10)
    p = get_test_parameter_set(1)[0]
    # complex
    jmc = mp.Value("i", 0)
    jmm = mp.Value("i", 0)
    odtbrain.backpropagate_3d_tilted(sino, angles, padval=0,
                                     dtype=float,
                                     count=jmc,
                                     max_count=jmm,
                                     **p)
    assert jmc.value == jmm.value
    assert jmc.value != 0


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
